"""User command functions for the mudpy engine."""

# Copyright (c) 2004-2020 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import random
import re
import traceback
import unicodedata

import mudpy


def chat(actor, parameters):
    """Toggle chat mode."""
    mode = actor.get("mode")
    if not mode:
        actor.set("mode", "chat")
        actor.send("Entering chat mode (use $(grn)!chat$(nrm) to exit).")
    elif mode == "chat":
        actor.remove_facet("mode")
        actor.send("Exiting chat mode.")
    else:
        actor.send("Sorry, but you're already busy with something else!")
    return True


def create(actor, parameters):
    """Create an element if it does not exist."""
    if not parameters:
        message = "You must at least specify an element to create."
    elif not actor.owner:
        message = ""
    else:
        arguments = parameters.split()
        if len(arguments) == 1:
            arguments.append("")
        if len(arguments) == 2:
            element, filename = arguments
            if element in actor.universe.contents:
                message = 'The "' + element + '" element already exists.'
            else:
                message = ('You create "' +
                           element + '" within the universe.')
                logline = actor.owner.account.get(
                    "name"
                ) + " created an element: " + element
                if filename:
                    logline += " in file " + filename
                    if filename not in actor.universe.files:
                        message += (
                            ' Warning: "' + filename + '" is not yet '
                            "included in any other file and will not be read "
                            "on startup unless this is remedied.")
                mudpy.misc.Element(element, actor.universe, filename)
                mudpy.misc.log(logline, 6)
        elif len(arguments) > 2:
            message = "You can only specify an element and a filename."
    actor.send(message)
    return True


def delete(actor, parameters):
    """Delete a facet from an element."""
    if not parameters:
        message = "You must specify an element and a facet."
    else:
        arguments = parameters.split(" ")
        if len(arguments) == 1:
            message = ('What facet of element "' + arguments[0]
                       + '" would you like to delete?')
        elif len(arguments) != 2:
            message = "You may only specify an element and a facet."
        else:
            element, facet = arguments
            if element not in actor.universe.contents:
                message = 'The "' + element + '" element does not exist.'
            elif facet not in actor.universe.contents[element].facets():
                message = ('The "' + element + '" element has no "' + facet
                           + '" facet.')
            else:
                actor.universe.contents[element].remove_facet(facet)
                message = ('You have successfully deleted the "' + facet
                           + '" facet of element "' + element
                           + '". Try "show element ' +
                           element + '" for verification.')
    actor.send(message)
    return True


def destroy(actor, parameters):
    """Destroy an element if it exists."""
    if actor.owner:
        if not parameters:
            message = "You must specify an element to destroy."
        else:
            if parameters not in actor.universe.contents:
                message = 'The "' + parameters + '" element does not exist.'
            else:
                actor.universe.contents[parameters].destroy()
                message = ('You destroy "' + parameters
                           + '" within the universe.')
                mudpy.misc.log(
                    actor.owner.account.get(
                        "name"
                    ) + " destroyed an element: " + parameters,
                    6
                )
        actor.send(message)
    return True


def error(actor, input_data):
    """Generic error for an unrecognized command word."""

    # 90% of the time use a generic error
    # Allow the random.randrange() call in bandit since it's not used for
    # security/cryptographic purposes
    if random.randrange(10):  # nosec
        message = '''I'm not sure what "''' + input_data + '''" means...'''

    # 10% of the time use the classic diku error
    else:
        message = "Arglebargle, glop-glyf!?!"

    # try to send the error message, and log if we can't
    try:
        actor.send(message)
    except Exception:
        mudpy.misc.log(
            'Sending a command error to user %s raised exception...\n%s' % (
                actor.owner.account.get("name"), traceback.format_exc()))
    return True


def evaluate(actor, parameters):
    """Evaluate a Python expression."""

    if not parameters:
        message = "You need to supply a Python expression."
    elif "__" in parameters:
        message = "Double-underscores (__) are not allowed in expressions."
    elif "lambda" in parameters:
        message = "Lambda functions are not allowed in expressions."
    else:
        # Strictly limit the allowed builtins and modules
        eval_globals = {"__builtins__": dict()}
        for allowed in ("dir", "globals", "len", "locals"):
            eval_globals["__builtins__"][allowed] = __builtins__[allowed]
        eval_globals["mudpy"] = mudpy
        eval_globals["universe"] = actor.universe
        try:
            # there is no other option than to use eval() for this, since
            # its purpose is to evaluate arbitrary expressions, so do what
            # we can to secure it and allow it for bandit analysis
            message = repr(eval(parameters, eval_globals))  # nosec
        except Exception as e:
            message = ("$(red)Your expression raised an exception...$(eol)"
                       "$(eol)$(bld)%s$(nrm)" % e)
    actor.send(message)
    return True


def halt(actor, parameters):
    """Halt the world."""
    if actor.owner:

        # see if there's a message or use a generic one
        if parameters:
            message = "Halting: " + parameters
        else:
            message = "User " + actor.owner.account.get(
                "name"
            ) + " halted the world."

        # let everyone know
        mudpy.misc.broadcast(message, add_prompt=False)
        mudpy.misc.log(message, 8)

        # set a flag to terminate the world
        actor.universe.terminate_flag = True
    return True


def help(actor, parameters):
    """List available commands and provide help for commands."""

    # did the user ask for help on a specific command word?
    if parameters and actor.owner:

        # is the command word one for which we have data?
        command = mudpy.misc.find_command(parameters)

        # only for allowed commands
        if actor.can_run(command):

            # add a description if provided
            description = command.get("description")
            if not description:
                description = "(no short description provided)"
            if command.is_restricted():
                output = "$(red)"
            else:
                output = "$(grn)"
            output = "%s%s$(nrm) - %s$(eol)$(eol)" % (
                output, command.subkey, description)

            # add the help text if provided
            help_text = command.get("help")
            if not help_text:
                help_text = "No help is provided for this command."
            output += help_text

            # list related commands
            see_also = command.get("see_also")
            if see_also:
                really_see_also = ""
                for item in see_also:
                    if item in actor.universe.groups["command"]:
                        command = actor.universe.groups["command"][item]
                        if actor.can_run(command):
                            if really_see_also:
                                really_see_also += ", "
                            if command.is_restricted():
                                really_see_also += "$(red)"
                            else:
                                really_see_also += "$(grn)"
                            really_see_also += item + "$(nrm)"
                if really_see_also:
                    output += "$(eol)$(eol)See also: " + really_see_also

        # no data for the requested command word
        else:
            output = "That is not an available command."

    # no specific command word was indicated
    else:

        # preamble text
        output = ("These are the commands available to you [brackets indicate "
                  "optional portion]:$(eol)$(eol)")

        # list command names in alphabetical order
        for command_name, command in sorted(
                actor.universe.groups["command"].items()):

            # skip over disallowed commands
            if actor.can_run(command):

                # start incrementing substrings
                for position in range(1, len(command_name) + 1):

                    # we've found our shortest possible abbreviation
                    candidate = mudpy.misc.find_command(
                            command_name[:position])
                    try:
                        if candidate.subkey == command_name:
                            break
                    except AttributeError:
                        pass

                # use square brackets to indicate optional part of command name
                if position < len(command_name):
                    abbrev = "%s[%s]" % (
                        command_name[:position], command_name[position:])
                else:
                    abbrev = command_name

                # supply a useful default if the short description is missing
                description = command.get(
                    "description", "(no short description provided)")

                # administrative command names are in red, others in green
                if command.is_restricted():
                    color = "red"
                else:
                    color = "grn"

                # format the entry for this command
                output = "%s   $(%s)%s$(nrm) - %s$(eol)" % (
                    output, color, abbrev, description)

        # add a footer with instructions on getting additional information
        output = ('%s $(eol)Enter "help COMMAND" for help on a command named '
                  '"COMMAND".' % output)

    # send the accumulated output to the user
    actor.send(output)
    return True


def look(actor, parameters):
    """Look around."""
    if parameters:
        actor.send("You can't look at or in anything yet.")
    else:
        actor.look_at(actor.get("location"))
    return True


def move(actor, parameters):
    """Move the avatar in a given direction."""
    for portal in sorted(
            actor.universe.contents[actor.get("location")].portals()):
        if portal.startswith(parameters):
            actor.move_direction(portal)
            return portal
    actor.send("You cannot go that way.")
    return True


def preferences(actor, parameters):
    """List, view and change actor preferences."""

    # Escape replacement macros in preferences
    parameters = mudpy.misc.escape_macros(parameters)

    message = ""
    arguments = parameters.split()
    allowed_prefs = set()
    base_prefs = []
    user_config = actor.universe.contents.get("mudpy.user")
    if user_config:
        base_prefs = user_config.get("pref_allow", [])
        allowed_prefs.update(base_prefs)
        if actor.owner.account.get("administrator"):
            allowed_prefs.update(user_config.get("pref_admin", []))
    if not arguments:
        message += "These are your current preferences:"

        # color-code base and admin prefs
        for pref in sorted(allowed_prefs):
            if pref in base_prefs:
                color = "grn"
            else:
                color = "red"
            message += ("$(eol)   $(%s)%s$(nrm) - %s" % (
                color, pref, actor.owner.account.get(pref, "<not set>")))

    elif arguments[0] not in allowed_prefs:
        message += (
            'Preference "%s" does not exist. Try the `preferences` command by '
            "itself for a list of valid preferences." % arguments[0])
    elif len(arguments) == 1:
        message += "%s" % actor.owner.account.get(arguments[0], "<not set>")
    else:
        pref = arguments[0]
        value = " ".join(arguments[1:])
        try:
            actor.owner.account.set(pref, value)
            message += 'Preference "%s" set to "%s".' % (pref, value)
        except ValueError:
            message = (
                'Preference "%s" cannot be set to type "%s".' % (
                    pref, type(value)))
    actor.send(message)
    return True


def quit(actor, parameters):
    """Leave the world and go back to the main menu."""
    if actor.owner:
        actor.owner.state = "main_utility"
        actor.owner.deactivate_avatar()
    return True


def reload(actor, parameters):
    """Reload all code modules, configs and data."""
    if actor.owner:

        # let the user know and log
        actor.send("Reloading all code modules, configs and data.")
        mudpy.misc.log(
            "User " +
            actor.owner.account.get("name") + " reloaded the world.",
            6
        )

        # set a flag to reload
        actor.universe.reload_flag = True
    return True


def say(actor, parameters):
    """Speak to others in the same area."""

    # check for replacement macros and escape them
    parameters = mudpy.misc.escape_macros(parameters)

    # if the message is wrapped in quotes, remove them and leave contents
    # intact
    if parameters.startswith('"') and parameters.endswith('"'):
        message = parameters[1:-1]
        literal = True

    # otherwise, get rid of stray quote marks on the ends of the message
    else:
        message = parameters.strip('''"'`''')
        literal = False

    # the user entered a message
    if message:

        # match the punctuation used, if any, to an action
        if "mudpy.linguistic" in actor.universe.contents:
            actions = actor.universe.contents[
                "mudpy.linguistic"].get("actions", {})
            default_punctuation = (actor.universe.contents[
                "mudpy.linguistic"].get("default_punctuation", "."))
        else:
            actions = {}
            default_punctuation = "."
        action = ""

        # reverse sort punctuation options so the longest match wins
        for mark in sorted(actions.keys(), reverse=True):
            if not literal and message.endswith(mark):
                action = actions[mark]
                break

        # add punctuation if needed
        if not action:
            action = actions[default_punctuation]
            if message and not (
               literal or unicodedata.category(message[-1]) == "Po"
               ):
                message += default_punctuation

        # failsafe checks to avoid unwanted reformatting and null strings
        if message and not literal:

            # decapitalize the first letter to improve matching
            message = message[0].lower() + message[1:]

            # iterate over all words in message, replacing typos
            if "mudpy.linguistic" in actor.universe.contents:
                typos = actor.universe.contents[
                    "mudpy.linguistic"].get("typos", {})
            else:
                typos = {}
            words = message.split()
            for index in range(len(words)):
                word = words[index]
                while unicodedata.category(word[0]) == "Po":
                    word = word[1:]
                while unicodedata.category(word[-1]) == "Po":
                    word = word[:-1]
                if word in typos.keys():
                    words[index] = words[index].replace(word, typos[word])
            message = " ".join(words)

            # capitalize the first letter
            message = message[0].upper() + message[1:]

    # tell the area
    if message:
        actor.echo_to_location(
            actor.get("name") + " " + action + 's, "' + message + '"'
        )
        actor.send("You " + action + ', "' + message + '"')

    # there was no message
    else:
        actor.send("What do you want to say?")
    return True


def c_set(actor, parameters):
    """Set a facet of an element."""
    if not parameters:
        message = "You must specify an element, a facet and a value."
    else:
        arguments = parameters.split(" ", 2)
        if len(arguments) == 1:
            message = ('What facet of element "' + arguments[0]
                       + '" would you like to set?')
        elif len(arguments) == 2:
            message = ('What value would you like to set for the "' +
                       arguments[1] + '" facet of the "' + arguments[0]
                       + '" element?')
        else:
            element, facet, value = arguments
            if element not in actor.universe.contents:
                message = 'The "' + element + '" element does not exist.'
            else:
                try:
                    actor.universe.contents[element].set(facet, value)
                except PermissionError:
                    message = ('The "%s" element is kept in read-only file '
                               '"%s" and cannot be altered.' %
                               (element, actor.universe.contents[
                                        element].origin.source))
                except ValueError:
                    message = ('Value "%s" of type "%s" cannot be coerced '
                               'to the correct datatype for facet "%s".' %
                               (value, type(value), facet))
                else:
                    message = ('You have successfully (re)set the "' + facet
                               + '" facet of element "' + element
                               + '". Try "show element ' +
                               element + '" for verification.')
    actor.send(message)
    return True


def show(actor, parameters):
    """Show program data."""
    message = ""
    arguments = parameters.split()
    if not parameters:
        message = "What do you want to show?"
    elif arguments[0] == "version":
        message = repr(actor.universe.versions)
    elif arguments[0] == "time":
        message = "%s increments elapsed since the world was created." % (
            str(actor.universe.groups["internal"]["counters"].get("elapsed")))
    elif arguments[0] == "groups":
        message = "These are the element groups:$(eol)"
        groups = list(actor.universe.groups.keys())
        groups.sort()
        for group in groups:
            message += "$(eol)   $(grn)" + group + "$(nrm)"
    elif arguments[0] == "files":
        message = "These are the current files containing the universe:$(eol)"
        filenames = sorted(actor.universe.files)
        for filename in filenames:
            if actor.universe.files[filename].is_writeable():
                status = "rw"
            else:
                status = "ro"
            message += ("$(eol)   $(red)(%s) $(grn)%s$(nrm)" %
                        (status, filename))
            if actor.universe.files[filename].flags:
                message += (" $(yel)[%s]$(nrm)" %
                            ",".join(actor.universe.files[filename].flags))
    elif arguments[0] == "group":
        if len(arguments) != 2:
            message = "You must specify one group."
        elif arguments[1] in actor.universe.groups:
            message = ('These are the elements in the "' + arguments[1]
                       + '" group:$(eol)')
            elements = [
                (
                    actor.universe.groups[arguments[1]][x].key
                ) for x in actor.universe.groups[arguments[1]].keys()
            ]
            elements.sort()
            for element in elements:
                message += "$(eol)   $(grn)" + element + "$(nrm)"
        else:
            message = 'Group "' + arguments[1] + '" does not exist.'
    elif arguments[0] == "file":
        if len(arguments) != 2:
            message = "You must specify one file."
        elif arguments[1] in actor.universe.files:
            message = ('These are the nodes in the "' + arguments[1]
                       + '" file:$(eol)')
            elements = sorted(actor.universe.files[arguments[1]].data)
            for element in elements:
                message += "$(eol)   $(grn)" + element + "$(nrm)"
        else:
            message = 'File "%s" does not exist.' % arguments[1]
    elif arguments[0] == "element":
        if len(arguments) != 2:
            message = "You must specify one element."
        elif arguments[1].strip(".") in actor.universe.contents:
            element = actor.universe.contents[arguments[1].strip(".")]
            message = ('These are the properties of the "' + arguments[1]
                       + '" element (in "' + element.origin.source
                       + '"):$(eol)')
            facets = element.facets()
            for facet in sorted(facets):
                message += ("$(eol)   $(grn)%s: $(red)%s$(nrm)" %
                            (facet, str(facets[facet])))
        else:
            message = 'Element "' + arguments[1] + '" does not exist.'
    elif arguments[0] == "log":
        if len(arguments) == 4:
            if re.match(r"^\d+$", arguments[3]) and int(arguments[3]) >= 0:
                stop = int(arguments[3])
            else:
                stop = -1
        else:
            stop = 0
        if len(arguments) >= 3:
            if re.match(r"^\d+$", arguments[2]) and int(arguments[2]) > 0:
                start = int(arguments[2])
            else:
                start = -1
        else:
            start = 10
        if len(arguments) >= 2:
            if (re.match(r"^\d+$", arguments[1])
                    and 0 <= int(arguments[1]) <= 9):
                level = int(arguments[1])
            else:
                level = -1
        elif 0 <= actor.owner.account.get("loglevel", 0) <= 9:
            level = actor.owner.account.get("loglevel", 0)
        else:
            level = 1
        if level > -1 and start > -1 and stop > -1:
            message = mudpy.misc.get_loglines(level, start, stop)
        else:
            message = ("When specified, level must be 0-9 (default 1), "
                       "start and stop must be >=1 (default 10 and 1).")
    else:
        message = '''I don't know what "''' + parameters + '" is.'
    actor.send(message)
    return True



"""Core modules package for the mudpy engine."""

# Copyright (c) 2004-2019 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import importlib

import mudpy


def load():
    """Import/reload some modules (be careful, as this can result in loops)."""

    # hard-coded fallback list of modules expected in this package
    # TODO(fungi) remove this once Python 3.6 is no longer supported
    fallback_modules = [
            "command",
            "daemon",
            "data",
            "menu",
            "misc",
            "password",
            "telnet",
            "version",
            ]
    try:
        # dynamically build module list from package contents (this only works
        # in Python 3.7 and later, hence the try/except)
        modules = []
        for module in mudpy.__loader__.contents():

            if (
                    # make sure it's a module file, not a directory
                    module.endswith('.py')

                    # don't include this file, we're inside it
                    and module != '__init__.py'):

                # trim off the .py file extension
                modules.append(module[:-3])

        # make sure the fallback list is kept up to date with package contents
        if fallback_modules != sorted(modules):
            raise Exception("Fallback module list is incomplete")

    except AttributeError:
        modules = fallback_modules

    # iterate over the list of module files included in the package
    for module in modules:

        # attempt to reload the module, assuming it was probably imported
        # earlier
        try:
            importlib.reload(getattr(mudpy, module))

        # must not have been, so import it now
        except AttributeError:
            importlib.import_module("mudpy.%s" % module)


load()



"""Password hashing functions and constants for the mudpy engine."""

# Copyright (c) 2004-2019 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import passlib.context

_CONTEXT = passlib.context.CryptContext(
    default="pbkdf2_sha512", pbkdf2_sha512__default_rounds=1000,
    schemes=["pbkdf2_sha512"])


def create(password):
    return _CONTEXT.hash(password)


def verify(password, encoded_hash):
    return _CONTEXT.verify(password, encoded_hash)



# Copyright (c) 2004-2021 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import os
import pathlib
import re
import shutil
import subprocess
import sys
import telnetlib
import time

import yaml

pidfile = "var/mudpy.pid"

test_account0_setup = (
    (0, "Identify yourself:", "luser0"),
    (0, "Enter your choice:", "n"),
    (0, 'Enter a new password for "luser0":', "Test123"),
    (0, "Enter the same new password again:", "Test123"),
    (0, r"What would you like to do\?", "c"),
    (0, "Pick a birth gender for your new avatar:", "f"),
    (0, "Choose a name for her:", "1"),
    (0, r"What would you like to do\?", "a"),
    (0, r"Whom would you like to awaken\?", ""),
)

test_account1_setup = (
    (1, "Identify yourself:", "luser1"),
    (1, "Enter your choice:", "n"),
    (1, 'Enter a new password for "luser1":', "Test456"),
    (1, "Enter the same new password again:", "Test456"),
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "1"),
    (1, r"What would you like to do\?", "a"),
    (1, r"Whom would you like to awaken\?", ""),
)

test_actor_appears = (
    (0, r"You suddenly realize that .* is here\.", ""),
)

test_explicit_punctuation = (
    (0, "> ", "say Hello there!"),
    (0, r'You exclaim, "Hello there\!"', ""),
    (1, r'exclaims, "Hello there\!"', "say And you are?"),
    (1, r'You ask, "And you are\?"', ""),
    (0, r'asks, "And you are\?"', "say I'm me, of course."),
    (0, r'''You say, "I'm me, of course\."''', ""),
    (1, r'''says, "I'm me, of course\."''', "say I wouldn't be so sure..."),
    (1, r'''You muse, "I wouldn't be so sure\.\.\."''', ""),
    (0, r'''muses, "I wouldn't be so sure\.\.\."''', "say You mean,"),
    (0, 'You begin, "You mean,"', ""),
    (1, 'begins, "You mean,"', "say I know-"),
    (1, 'You begin, "I know-"', ""),
    (0, 'begins, "I know-"', "say Don't interrupt:"),
    (0, r'''You begin, "Don't interrupt:"''', ""),
    (1, r'''begins, "Don't interrupt:"''', "say I wasn't interrupting;"),
    (1, r'''You begin, "I wasn't interrupting;"''', ""),
    (0, r'''begins, "I wasn't interrupting;"''', ""),
)

test_implicit_punctuation = (
    (0, '> ', "say Whatever"),
    (0, r'You say, "Whatever\."', ""),
    (1, r'says, "Whatever\."', ""),
)

test_typo_replacement = (
    (1, '> ', "say That's what i think."),
    (1, r'''You say, "That's what I think\."''', ""),
    (0, r'''says, "That's what I think\."''', "say You know what i'd like."),
    (0, r'''You say, "You know what I'd like\."''', ""),
    (1, r'''says, "You know what I'd like\."''', "say Then i'll tell you."),
    (1, r'''You say, "Then I'll tell you\."''', ""),
    (0, r'''says, "Then I'll tell you\."''', "say Now i'm ready."),
    (0, r'''You say, "Now I'm ready\."''', ""),
    (1, r'''says, "Now I'm ready\."''', "say That's teh idea."),
    (1, r'''You say, "That's the idea\."''', ""),
    (0, r'''says, "That's the idea\."''', "say It's what theyre saying."),
    (0, r'''You say, "It's what they're saying\."''', ""),
    (1, r'''says, "It's what they're saying\."''', "say Well, youre right."),
    (1, r'''You say, "Well, you're right\."''', ""),
    (0, r'''says, "Well, you're right\."''', ""),
)

test_sentence_capitalization = (
    (0, "> ", "say this sentence should start with a capital T."),
    (0, 'You say, "This sentence', ""),
    (1, 'says, "This sentence', ""),
)

test_chat_mode = (
    (1, '> ', "chat"),
    (1, r'(?s)Entering chat mode .*> \(chat\) ', "Feeling chatty."),
    (1, r'You say, "Feeling chatty\."', "!chat"),
    (0, r'says, "Feeling chatty\."', ""),
    (1, '> ', "say Now less chatty."),
    (1, r'You say, "Now less chatty\."', ""),
    (0, r'says, "Now less chatty\."', ""),
)

test_wrapping = (
    (0, '> ', "say " + 100 * "o"),
    (1, r'says,\r\n"O[o]+\."', ""),
)

test_forbid_ansi_input = (
    (0, '> ', "say \x1b[35mfoo\x1b[0m"),
    (1, r'says, "\[35mfoo\[0m\."', ""),
)

test_escape_macros = (
    (0, '> ', "say $(red)bar$(nrm)"),
    (1, r'says, "\$\(red\)bar\$\(nrm\)\."', ""),
)

test_movement = (
    (0, "> ", "move north"),
    (0, r"You exit to the north\.", ""),
    (1, r"exits to the north\.", "move north"),
    (0, r"arrives from the south\.", "move south"),
    (0, r"You exit to the south\.", ""),
    (1, r"exits to the south\.", "move south"),
    (0, r"arrives from the north\.", "move east"),
    (0, r"You exit to the east\.", ""),
    (1, r"exits to the east\.", "move east"),
    (0, r"arrives from the west\.", "move west"),
    (0, r"You exit to the west\.", ""),
    (1, r"exits to the west\.", "move west"),
    (0, r"arrives from the east\.", "move up"),
    (0, r"You exit upward\.", ""),
    (1, r"exits upward\.", "move up"),
    (0, r"arrives from below\.", "move down"),
    (0, r"You exit downward\.", ""),
    (1, r"exits downward\.", "move down"),
    (0, r"arrives from above\.", ""),
)

test_actor_disappears = (
    (1, "> ", "quit"),
    (0, r"You suddenly wonder where .* went\.", ""),
)

test_abort_avatar_deletion = (
    (1, r"What would you like to do\?", "d"),
    (1, r"Whom would you like to delete\?", ""),
    (1, r"delete an unwanted avatar.*What would you like to do\?", ""),
)

test_avatar_creation_limit = (
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "3"),
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "4"),
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "7"),
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "5"),
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "2"),
    (1, r"What would you like to do\?", "c"),
    (1, "Pick a birth gender for your new avatar:", "m"),
    (1, "Choose a name for him:", "6"),
    (1, r"What would you like to do\?", "c"),
    (1, r"That is not a valid choice\.\.\.", ""),
)

test_avatar_deletion = (
    (1, r"What would you like to do\?", "d"),
    (1, r"Whom would you like to delete\?", "1"),
    (1, r"create a new avatar.*What would you like to do\?", ""),
)

test_abort_account_deletion = (
    (1, r"What would you like to do\?", "p"),
    (1, r"permanently delete your account\?", ""),
    (1, r"What would you like to do\?", ""),
)

test_account_deletion = (
    (1, r"What would you like to do\?", "p"),
    (1, r"permanently delete your account\?", "y"),
    (1, r"Disconnecting\.\.\.", ""),
)

test_admin_setup = (
    (2, "Identify yourself:", "admin"),
    (2, "Enter your choice:", "n"),
    (2, 'Enter a new password for "admin":', "Test789"),
    (2, "Enter the same new password again:", "Test789"),
    (2, r"What would you like to do\?", "c"),
    (2, "Pick a birth gender for your new avatar:", "m"),
    (2, "Choose a name for him:", "1"),
    (2, r"What would you like to do\?", "a"),
    (2, r"Whom would you like to awaken\?", ""),
)

test_preferences = (
    (0, "> ", "preferences"),
    (0, r"\[32mprompt\x1b\[0m - <not set>.*> ", "preferences prompt $(foo)"),
    (0, r"\$\(foo\) ", "preferences prompt"),
    (0, r"\$\(foo\).*\$\(foo\) ", "preferences prompt $(time)>"),
    (0, "[0-9]> ", "preferences loglevel 0"),
    (0, "does not exist.*> ", "preferences prompt >"),
    (2, "> ", "preferences loglevel 0"),
    (2, "> ", "preferences"),
    (2, r"\[31mloglevel\x1b\[0m - 0.*> ", "preferences loglevel zero"),
    (2, r'''cannot be set to type "<class 'str'>"\..*> ''', ""),
)

test_crlf_eol = (
    # Send a CR+LF at the end of the line instead of the default CR+NUL,
    # to make sure they're treated the same
    (2, "> ", b"say I use CR+LF as my EOL, not CR+NUL.\r\n"),
    (2, r'You say, "I use CR\+LF as my EOL, not CR\+NUL\.".*> ', ""),
)

test_telnet_iac = (
    # Send a double (escaped) IAC byte within other text, which should get
    # unescaped and deduplicated to a single \xff in the buffer and then
    # the line of input discarded as a non-ASCII sequence
    (2, "> ", b"say argle\xff\xffbargle\r\0"),
    (2, r"Non-ASCII characters from admin: b'say.*argle\\xffbargle'.*> ", ""),
)

test_telnet_unknown_command = (
    # Send an unsupported negotiation command #127 which should get filtered
    # from the line of input
    (2, "> ", b"say glop\xff\x7fglyf\r\0"),
    (2, r'Ignored unknown command 127 from admin\..*"Glopglyf\.".*> ', ""),
)

test_telnet_unknown_option = (
    # Send an unassigned negotiation option #127 which should get logged
    (2, "> ", b"\xff\xfe\x7f\r\0"),
    (2, r'''Received "don't 127" from admin\..*> ''', ""),
)

test_admin_restriction = (
    (0, "> ", "help halt"),
    (0, r"That is not an available command\.", "halt"),
    (0, '(not sure what "halt" means|Arglebargle, glop-glyf)', ""),
)

test_admin_help = (
    (2, "> ", "help"),
    (2, r"halt.*Shut down the world\.", "help halt"),
    (2, "This will save all active accounts", ""),
)

test_help = (
    (0, "> ", "help say"),
    (0, r"See also: .*chat.*> ", ""),
)

test_abbrev = (
    (0, "> ", "h"),
    (0, r"h\[elp\].*m\[ove\].*> ", "he mo"),
    (0, r"Move in a specific direction\..*> ", "mov nor"),
    (0, r"You exit to the north\..*> ", "m s"),
    (0, r"You exit to the south\..*> ", ""),
)

test_reload = (
    (2, "> ", "reload"),
    (2, r"Reloading all code modules, configs and data\."
        r".* User admin reloaded the world\.",
     "show element account.admin"),
    (2, r'These are the properties of the "account\.admin" element.*'
        r'  \x1b\[32mpasshash:\r\n\x1b\[31m\$.*> ', ""),
)

test_set_facet = (
    (2, "> ", "set actor.avatar_admin_0 gender female"),
    (2, r'You have successfully \(re\)set the "gender" facet of element', ""),
)

test_set_refused = (
    (2, "> ", "set mudpy.limit password_tries 10"),
    (2, r'The "mudpy\.limit" element is kept in read-only file', ""),
)

test_show_version = (
    (2, "> ", "show version"),
    (2, r"Running mudpy .* on .* Python 3.*with.*pyyaml.*> ", ""),
)

test_show_time = (
    (2, "> ", "show time"),
    (2, r"\r\n[0-9]+ increments elapsed.*> ", ""),
)

test_show_files = (
    (2, "> ", "show files"),
    (2, r'These are the current files containing the universe:.*'
        r'  \x1b\[31m\(rw\) \x1b\[32m/.*/account\.yaml\x1b\[0m'
        r' \x1b\[33m\[private\]\x1b\[0m.*> ', ""),
)

test_show_file = (
    (2, "> ", "show file %s" %
        os.path.join(os.getcwd(), "data/internal.yaml")),
    (2, r'These are the nodes in the.*file:.*internal\.counters.*> ', ""),
)

test_show_groups = (
    (2, "> ", "show groups"),
    (2, r'These are the element groups:.*'
        r'  \x1b\[32maccount\x1b\[0m.*> ', ""),
)

test_show_group = (
    (2, "> ", "show group account"),
    (2, r'These are the elements in the "account" group:.*'
        r'  \x1b\[32maccount\.admin\x1b\[0m.*> ', ""),
)

test_show_element = (
    (2, "> ", "show element mudpy.limit"),
    (2, r'These are the properties of the "mudpy\.limit" element.*'
        r'  \x1b\[32mpassword_tries: \x1b\[31m3.*> ',
     "show element actor.avatar_admin_0"),
    (2, r'These are the properties of the "actor\.avatar_admin_0" element.*'
        r'  \x1b\[32mgender: \x1b\[31mfemale.*> ', ""),
)

test_evaluate = (
    (2, "> ", "evaluate 12345*67890"),
    (2, r"\r\n838102050\r\n.*> ", "evaluate 1/0"),
    (2, "Your expression raised an exception.*division by zero.*> ",
     "evaluate mudpy"),
    (2, "<module 'mudpy' from.*> ", "evaluate re"),
    (2, "Your expression raised an exception.*name 're' is not defined.*> ",
     "evaluate universe"),
    (2, r"<mudpy\.misc\.Universe object at 0x.*> ", "evaluate actor"),
    (2, "Your expression raised an exception.*name 'actor' is not defined.*> ",
        "evaluate dir(mudpy)"),
    (2, "__builtins__.*> ", "evaluate mudpy.__builtins__.open"),
    (2, "not allowed.*> ", "evaluate (lambda x: x + 1)(2)"),
    (2, "not allowed.*> ", ""),
)

test_debug_restricted = (
    (0, "> ", "help evaluate"),
    (0, r"That is not an available command\.", "evaluate"),
    (0, '(not sure what "evaluate" means|Arglebargle, glop-glyf)', ""),
)

test_debug_disabled = (
    (2, "> ", "help evaluate"),
    (2, r"That is not an available command\.", "evaluate"),
    (2, '(not sure what "evaluate" means|Arglebargle, glop-glyf)', ""),
)

test_show_log = (
    (2, "> ", "show log"),
    (2, r"There are [0-9]+ log lines in memory and [0-9]+ at or above level "
        r"[0-9]+\. The matching.*from [0-9]+ to [0-9]+ are:", ""),
)

test_custom_loglevel = (
    (2, "> ", "set account.admin loglevel 2"),
    (2, "You have successfully .*> ", "show log"),
    (2, r"There are [0-9]+ log lines in memory and [0-9]+ at or above level "
        r"[0-9]+\. The matching.*from [0-9]+ to [0-9]+ are:", ""),
)

test_invalid_loglevel = (
    (2, "> ", "set account.admin loglevel two"),
    (2, r'''Value "two" of type "<class 'str'>" cannot be coerced .*> ''', ""),
)

test_log_no_errors = (
    (2, "> ", "show log 7"),
    (2, r"None of the [0-9]+ lines in memory matches your request\.", ""),
)

final_cleanup = (
    (0, "> ", "quit"),
    (0, r"What would you like to do\?", "p"),
    (0, r"permanently delete your account\?", "y"),
    (0, r"Disconnecting\.\.\.", ""),
    (2, "> ", "quit"),
    (2, r"What would you like to do\?", "p"),
    (2, r"permanently delete your account\?", "y"),
    (2, r"Disconnecting\.\.\.", ""),
)

dialogue = {
    test_account0_setup: "first account setup",
    test_account1_setup: "second account setup",
    test_actor_appears: "actor spontaneous appearance",
    test_explicit_punctuation: "explicit punctuation",
    test_implicit_punctuation: "implicit punctuation",
    test_typo_replacement: "typo replacement",
    test_sentence_capitalization: "sentence capitalization",
    test_chat_mode: "chat mode",
    test_wrapping: "wrapping",
    test_forbid_ansi_input: "raw escape input is filtered",
    test_escape_macros: "replacement macros are escaped",
    test_movement: "movement",
    test_actor_disappears: "actor spontaneous disappearance",
    test_abort_avatar_deletion: "abort avatar deletion",
    test_avatar_creation_limit: "avatar creation limit",
    test_avatar_deletion: "avatar deletion",
    test_abort_account_deletion: "abort account deletion",
    test_account_deletion: "account deletion",
    test_admin_setup: "admin account setup",
    test_preferences: "set and show preferences",
    test_crlf_eol: "send crlf from the client as eol",
    test_telnet_iac: "escape stray telnet iac bytes",
    test_telnet_unknown_command: "strip unknown telnet command",
    test_telnet_unknown_option: "log unknown telnet option",
    test_admin_restriction: "restricted admin commands",
    test_admin_help: "admin help",
    test_help: "help command",
    test_abbrev: "command abbreviation",
    test_reload: "reload",
    test_set_facet: "set facet",
    test_set_refused: "refuse altering read-only element",
    test_show_version: "show version and diagnostic info",
    test_show_time: "show elapsed world clock increments",
    test_show_files: "show a list of loaded files",
    test_show_file: "show nodes from a specific file",
    test_show_groups: "show groups",
    test_show_group: "show group",
    test_show_element: "show element",
    test_evaluate: "show results of python expressions",
    test_debug_restricted: "only admins can run debug commands",
    test_debug_disabled: "debugging commands only in debug mode",
    test_show_log: "show log",
    test_custom_loglevel: "custom loglevel",
    test_invalid_loglevel: "invalid loglevel",
    test_log_no_errors: "no errors logged",
    final_cleanup: "delete remaining accounts",
}

debug_tests = (
    test_evaluate,
)

nondebug_tests = (
    test_debug_disabled,
)


def start_service(config):
    # Clean up any previously run daemon which didn't terminate
    if os.path.exists(pidfile):
        with open(pidfile) as pidfd:
            pid = int(pidfd.read())
        try:
            # Stop the running service
            os.kill(pid, 15)
        except ProcessLookupError:
            # If there was no process, just remove the stale PID file
            os.remove(pidfile)
        # If there's a preexisting hung service, we can't proceed
        assert not os.path.exists(pidfile)

    # Clean up any previous test output
    for f in pathlib.Path(".").glob("capture_*.log"):
        os.remove(f)
    for d in ("data", "var"):
        shutil.rmtree(d, ignore_errors=True)

    # Start the service and wait for it to be ready for connections
    service = subprocess.Popen(("mudpy", config),
                               stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    return service


def stop_service(service):
    success = True

    # The no-op case when no service was started
    if service is None:
        return success

    # This handles when the service is running as a direct child process
    service.terminate()
    returncode = service.wait(10)
    if returncode != 0:
        tlog("\nERROR: Service exited with code %s." % returncode)
        success = False

    # This cleans up a daemonized and disassociated service
    if os.path.exists(pidfile):
        with open(pidfile) as pidfd:
            pid = int(pidfd.read())
        try:
            # Stop the running service
            os.kill(pid, 15)
            time.sleep(1)
        except ProcessLookupError:
            # If there was no process, just remove the stale PID file
            os.remove(pidfile)
        # The PID file didn't disappear, so we have a hung service
        if os.path.exists(pidfile):
            tlog("\nERROR: Hung daemon with PID %s." % pid)
            success = False

    # Log the contents of stdout and stderr, if any
    stdout, stderr = service.communicate()
    tlog("\nRecording stdout as capture_stdout.log.")
    with open("capture_stdout.log", "w") as serviceout:
        serviceout.write(stdout.decode("utf-8"))
    tlog("\nRecording stderr as capture_stderr.log.")
    with open("capture_stderr.log", "w") as serviceerr:
        serviceerr.write(stderr.decode("utf-8"))

    # Error if anything was written on stderr as this may indicate ignored
    # exceptions (e.g. ResourceWarning during garbage collection)
    if stderr:
        tlog("\nERROR: something was written to stderr, see "
             "capture_stderr.log for details.")
        success = False

    return success


def tlog(message, quiet=False):
    logfile = "capture_tests.log"
    with open(logfile, "a") as logfd:
        logfd.write(message + "\n")
    if not quiet:
        sys.stdout.write(message)
    return True


def option_callback(telnet_socket, command, option):
    if option == b'\x7f':
        # We use this unassigned option value as a canary, so short-circuit
        # any response to avoid endlessly looping
        pass
    elif command in (telnetlib.DO, telnetlib.DONT):
        telnet_socket.send(telnetlib.IAC + telnetlib.WONT + option)
    elif command in (telnetlib.WILL, telnetlib.WONT):
        telnet_socket.send(telnetlib.IAC + telnetlib.DONT + option)


def check_debug():
    if len(sys.argv) > 1:
        with open(sys.argv[1]) as config_fd:
            config = yaml.safe_load(config_fd)
        return config.get(".mudpy.limit.debug", False)
    return False


def connect_client(luser, service):
    # Try multiple times to connect, with an exponential backoff
    for retry in range(5):
        try:
            # Skipping the retry=0 case gives an immediate first attempt
            if retry:
                time.sleep((2 ** retry) / 10)
            luser.open("::1", 4000)
            # Attempt to poll the connection, closing if unusable
            try:
                luser.fill_rawq()
            except ConnectionResetError:
                luser.close()
                continue
            # Short-circuit if we get this far, connection is safe to use
            return luser
        except ConnectionRefusedError:
            continue
    else:
        # Connection retries have been exhausted, so give up
        tlog("\nERROR: Client could not connect.\n")
        stop_service(service)
        sys.exit(1)


def main():
    captures = ["", "", ""]
    lusers = [telnetlib.Telnet(), telnetlib.Telnet(), telnetlib.Telnet()]
    success = True
    start = time.time()
    service = None
    if len(sys.argv) > 1:
        # Start the service if a config file was provided on the command line
        service = start_service(sys.argv[1])
        if not service:
            tlog("\nERROR: Service did not start.\n")
            sys.exit(1)
    for luser in lusers:
        connect_client(luser, service)
        luser.set_option_negotiation_callback(option_callback)
    selected_dialogue = dict(dialogue)
    if check_debug():
        for test in nondebug_tests:
            del selected_dialogue[test]
    else:
        for test in debug_tests:
            del selected_dialogue[test]
    for test, description in selected_dialogue.items():
        tlog("\nTesting %s..." % description)
        test_start = time.time()
        for conversant, question, answer in test:
            tlog("luser%s waiting for: %s" % (conversant, question),
                 quiet=True)
            try:
                index, match, received = lusers[conversant].expect(
                    [re.compile(question.encode("utf-8"), flags=re.DOTALL)], 5)
                captures[conversant] += received.decode("utf-8")
            except (ConnectionResetError, EOFError):
                tlog("\nERROR: luser%s premature disconnection expecting:\n\n"
                     "%s\n\n"
                     "Check the end of capture_%s.log for received data."
                     % (conversant, question, conversant))
                success = False
                break
            try:
                captures[conversant] += lusers[
                    conversant].read_very_eager().decode("utf-8")
            except Exception:
                pass
            if index != 0:
                tlog("\nERROR: luser%s did not receive expected string:\n\n"
                     "%s\n\n"
                     "Check the end of capture_%s.log for received data."
                     % (conversant, question, conversant))
                success = False
                break
            if type(answer) is str:
                tlog("luser%s sending: %s" % (conversant, answer), quiet=True)
                lusers[conversant].write(("%s\r\0" % answer).encode("utf-8"))
                captures[conversant] += "%s\r\n" % answer
            elif type(answer) is bytes:
                tlog("luser%s sending raw bytes: %s" % (conversant, answer),
                     quiet=True)
                lusers[conversant].get_socket().send(answer)
                captures[conversant] += "!!!RAW BYTES: %s" % answer
            else:
                tlog("\nERROR: answer provided with unsupported type %s"
                     % type(answer))
                success = False
                break
        if not success:
            break
        tlog("Completed in %.3f seconds." % (time.time() - test_start))
    duration = time.time() - start
    for conversant in range(len(captures)):
        try:
            captures[conversant] += lusers[
                conversant].read_very_eager().decode("utf-8")
        except Exception:
            pass
        lusers[conversant].close()
        logfile = "capture_%s.log" % conversant
        tlog("\nRecording session %s as %s." % (conversant, logfile))
        log = open(logfile, "w")
        log.write(captures[conversant])
        log.close()
    if not stop_service(service):
        success = False
    tlog("\nRan %s tests in %.3f seconds." % (len(dialogue), duration))
    if success:
        tlog("\nSUCCESS\n")
    else:
        tlog("\nFAILURE\n")
        sys.exit(1)


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2004-2018 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

# core objects for the mudpy engine
import importlib
import sys

import mudpy


def main():

    # start it up
    mudpy.misc.setup()

    # loop indefinitely while the world is not flagged for termination or
    # there are still connected users
    while (not mudpy.misc.universe.terminate_flag or
           mudpy.misc.universe.userlist):

        # the world was flagged for a reload of all code/data
        if mudpy.misc.universe.reload_flag:
            importlib.reload(mudpy)
            mudpy.misc.reload_data()
            mudpy.misc.universe.reload_flag = False

        # do what needs to be done on each pulse
        mudpy.misc.on_pulse()

    # shut it all down
    mudpy.misc.finish()


if __name__ == '__main__':
    sys.exit(main())



"""Utility functions for menu operation."""

# Copyright (c) 2019 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import mudpy


def activate_avatar_action(user):
    """Activate the selected avatar in the activate_avatar state."""
    return user.activate_avatar_by_index(int(user.choice) - 1)


def activate_avatar_action_a(user):
    """Abort selection by doing nothing."""
    return True


def activate_avatar_create(user):
    """List available avatars as choices for the activate_avatar state."""
    return dict(
        [(str(x + 1), y) for x, y in enumerate(user.list_avatar_names())])


def checking_new_account_name_action_d(user):
    """Destroy the new account because the user asked to disconnect."""
    return user.account.destroy()


def checking_new_account_name_action_g(user):
    """Destroy the new account, the user asked to go back to name entry."""
    return user.account.destroy()


def choose_gender_action(user):
    """Set the avatar's gender to the selected value."""
    return user.avatar.set("gender", user.menu_choices[user.choice])


def choose_name_action(user):
    """Set the avatar's name to the selected value."""
    return user.avatar.set("name", user.menu_choices[user.choice])


def choose_name_create_1(user):
    """Provide a randomly-generated name as choice 1."""
    return mudpy.misc.random_name()


def choose_name_create_2(user):
    """Provide a randomly-generated name as choice 2."""
    return mudpy.misc.random_name()


def choose_name_create_3(user):
    """Provide a randomly-generated name as choice 3."""
    return mudpy.misc.random_name()


def choose_name_create_4(user):
    """Provide a randomly-generated name as choice 4."""
    return mudpy.misc.random_name()


def choose_name_create_5(user):
    """Provide a randomly-generated name as choice 5."""
    return mudpy.misc.random_name()


def choose_name_create_6(user):
    """Provide a randomly-generated name as choice 6."""
    return mudpy.misc.random_name()


def choose_name_create_7(user):
    """Provide a randomly-generated name as choice 7."""
    return mudpy.misc.random_name()


def delete_account_action_y(user):
    """Permanently delete the account and all avatars, as requested."""
    return user.destroy()


def delete_avatar_action(user):
    """Delete the selected avatar."""
    return user.delete_avatar(
        user.account.get("avatars")[int(user.choice) - 1])


def delete_avatar_action_a(user):
    """Abort avatar deletion."""
    return True


def delete_avatar_create(user):
    """List available avatars for possible deletion."""
    return dict(
        [(str(x + 1), y) for x, y in enumerate(user.list_avatar_names())])


def main_utility_action_c(user):
    """Create a new avatar."""
    return user.new_avatar()


def main_utility_demand_a(user):
    """Only include avatar activation if the account has avatars."""
    return user.account.get("avatars")


def main_utility_demand_c(user):
    """Only include avatar creation if avatar count is below the limit."""
    return (len(user.account.get("avatars")) <
            mudpy.misc.universe.contents["mudpy.limit"].get("avatars"))


def main_utility_demand_d(user):
    """Only include avatar deletion if the account has avatars."""
    return user.account.get("avatars")



"""Version and diagnostic information for the mudpy engine."""

# Copyright (c) 2018-2020 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import json
import sys


# TODO(fungi) Clean up once Python 3.8 is the oldest interpreter supported
try:
    import importlib.metadata
    use_importlib = True
except ModuleNotFoundError:
    import pkg_resources
    use_importlib = False


class VersionDetail:

    """Version detail for a Python package."""

    def __init__(self, package):
        if use_importlib:
            project_name = package.metadata.get('Name')
        else:
            project_name = package.project_name
        self.project_name = _normalize_project(project_name)
        version = package.version
        self.version_info = tuple(version.split('.'))

        # Build up a human-friendly version string for display purposes
        self.text = "%s %s" % (self.project_name, version)

        # Obtain Git commit ID from PBR metadata if present
        if use_importlib:
            dist = importlib.metadata.distribution(self.project_name)
        else:
            dist = pkg_resources.get_distribution(self.project_name)
        try:
            if use_importlib:
                pbr_metadata = dist.read_text("pbr.json")
            else:
                pbr_metadata = dist.get_metadata("pbr.json")
        except (IOError, KeyError):
            pbr_metadata = None
        if pbr_metadata:
            self.git_version = json.loads(pbr_metadata)["git_version"]
        else:
            self.git_version = None
        if self.git_version:
            self.text = "%s (%s)" % (self.text, self.git_version)

    def __repr__(self):
        return self.text


class Versions:

    """Tracks info on known Python package versions."""

    def __init__(self, project_name):
        # Normalize the supplied name
        project_name = _normalize_project(project_name)

        # Python info for convenience
        self.python_version = "%s Python %s" % (
            sys.platform, sys.version.split(" ")[0])

        # List of package names for this package's declared dependencies
        requirements = []
        if use_importlib:
            for req in importlib.metadata.distribution(project_name).requires:
                requirements.append(_normalize_project(req))
        else:
            for req in pkg_resources.get_distribution(project_name).requires():
                requirements.append(_normalize_project(req.project_name))

        # Accumulators for Python package versions
        self.dependencies = {}
        self.environment = {}

        # Loop over all installed packages
        if use_importlib:
            distributions = importlib.metadata.distributions()
        else:
            distributions = pkg_resources.working_set
        for package in distributions:
            version = VersionDetail(package)
            # Sort packages into the corresponding buckets
            if version.project_name in requirements:
                # This is a dependency
                self.dependencies[version.project_name] = version
            elif version.project_name == project_name:
                # This is our main package
                self.version = version
            else:
                # This may be a transitive dep, or merely installed
                self.environment[version.project_name] = version

        self.dependencies_text = ", ".join(
            sorted([x.text for x in self.dependencies.values()]))
        self.environment_text = ", ".join(
            sorted([x.text for x in self.environment.values()]))

    def __repr__(self):
        return "Running %s on %s with dependencies %s." % (
            self.version.text,
            self.python_version,
            self.dependencies_text,
            )


def _normalize_project(project_name):
    """Convenience function to normalize Python project names."""
    if use_importlib:
        return project_name.lower()
    else:
        return pkg_resources.safe_name(project_name).lower()



"""Miscellaneous functions for the mudpy engine."""

# Copyright (c) 2004-2021 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import codecs
import datetime
import os
import random
import re
import signal
import socket
import sys
import syslog
import time
import traceback
import unicodedata

import mudpy


class Element:

    """An element of the universe."""

    def __init__(self, key, universe, origin=None):
        """Set up a new element."""

        # keep track of our key name
        self.key = key

        # keep track of what universe it's loading into
        self.universe = universe

        # set of facet keys from the universe
        self.facethash = dict()

        # not owned by a user by default (used for avatars)
        self.owner = None

        # no contents in here by default
        self.contents = {}

        if self.key.find(".") > 0:
            self.group, self.subkey = self.key.split(".")[-2:]
        else:
            self.group = "other"
            self.subkey = self.key
        if self.group not in self.universe.groups:
            self.universe.groups[self.group] = {}

        # get an appropriate origin
        if not origin:
            self.universe.add_group(self.group)
            origin = self.universe.files[
                    self.universe.origins[self.group]["fallback"]]

        # record or reset a pointer to the origin file
        self.origin = self.universe.files[origin.source]

        # add or replace this element in the universe
        self.universe.contents[self.key] = self
        self.universe.groups[self.group][self.subkey] = self

    def reload(self):
        """Create a new element and replace this one."""
        args = (self.key, self.universe, self.origin)
        self.destroy()
        Element(*args)

    def destroy(self):
        """Remove an element from the universe and destroy it."""
        for facet in dict(self.facethash):
            self.remove_facet(facet)
        del self.universe.groups[self.group][self.subkey]
        del self.universe.contents[self.key]
        del self

    def facets(self):
        """Return a list of non-inherited facets for this element."""
        return self.facethash

    def has_facet(self, facet):
        """Return whether the non-inherited facet exists."""
        return facet in self.facets()

    def remove_facet(self, facet):
        """Remove a facet from the element."""
        if ".".join((self.key, facet)) in self.origin.data:
            del self.origin.data[".".join((self.key, facet))]
        if facet in self.facethash:
            del self.facethash[facet]
        self.origin.modified = True

    def ancestry(self):
        """Return a list of the element's inheritance lineage."""
        if self.has_facet("inherit"):
            ancestry = self.get("inherit")
            if not ancestry:
                ancestry = []
            for parent in ancestry[:]:
                ancestors = self.universe.contents[parent].ancestry()
                for ancestor in ancestors:
                    if ancestor not in ancestry:
                        ancestry.append(ancestor)
            return ancestry
        else:
            return []

    def get(self, facet, default=None):
        """Retrieve values."""
        if default is None:
            default = ""
        try:
            return self.origin.data[".".join((self.key, facet))]
        except (KeyError, TypeError):
            pass
        if self.has_facet("inherit"):
            for ancestor in self.ancestry():
                if self.universe.contents[ancestor].has_facet(facet):
                    return self.universe.contents[ancestor].get(facet)
        else:
            return default

    def set(self, facet, value):
        """Set values."""
        if not self.origin.is_writeable() and not self.universe.loading:
            # break if there is an attempt to update an element from a
            # read-only file, unless the universe is in the midst of loading
            # updated data from files
            raise PermissionError("Altering elements in read-only files is "
                                  "disallowed")
        # Coerce some values to appropriate data types
        # TODO(fungi) Move these to a separate validation mechanism
        if facet in ["loglevel"]:
            value = int(value)
        elif facet in ["administrator"]:
            value = bool(value)

        # The canonical node for this facet within its origin
        node = ".".join((self.key, facet))

        if node not in self.origin.data or self.origin.data[node] != value:
            # Be careful to only update the origin's contents when required,
            # since that affects whether the backing file gets written
            self.origin.data[node] = value
            self.origin.modified = True

        # Make sure this facet is included in the element's facets
        self.facethash[facet] = self.origin.data[node]

    def append(self, facet, value):
        """Append value to a list."""
        newlist = self.get(facet)
        if not newlist:
            newlist = []
        if type(newlist) is not list:
            newlist = list(newlist)
        newlist.append(value)
        self.set(facet, newlist)

    def send(
        self,
        message,
        eol="$(eol)",
        raw=False,
        flush=False,
        add_prompt=True,
        just_prompt=False,
        add_terminator=False,
        prepend_padding=True
    ):
        """Convenience method to pass messages to an owner."""
        if self.owner:
            self.owner.send(
                message,
                eol,
                raw,
                flush,
                add_prompt,
                just_prompt,
                add_terminator,
                prepend_padding
            )

    def is_restricted(self):
        """Boolean check whether command is administrative or debugging."""
        return bool(self.get("administrative") or self.get("debugging"))

    def is_admin(self):
        """Boolean check whether an actor is controlled by an admin owner."""
        return self.owner and self.owner.is_admin()

    def can_run(self, command):
        """Check if the user can run this command object."""

        # has to be in the commands group
        if command not in self.universe.groups["command"].values():
            return False

        # debugging commands are not allowed outside debug mode
        if command.get("debugging") and not self.universe.debug_mode():
            return False

        # avatars of administrators can run any command
        if self.is_admin():
            return True

        # everyone can run non-administrative commands
        if not command.is_restricted():
            return True

        # otherwise the command cannot be run by this actor
        return False

    def update_location(self):
        """Make sure the location's contents contain this element."""
        area = self.get("location")
        if area in self.universe.contents:
            self.universe.contents[area].contents[self.key] = self

    def clean_contents(self):
        """Make sure the element's contents aren't bogus."""
        for element in self.contents.values():
            if element.get("location") != self.key:
                del self.contents[element.key]

    def go_to(self, area):
        """Relocate the element to a specific area."""
        current = self.get("location")
        if current and self.key in self.universe.contents[current].contents:
            del universe.contents[current].contents[self.key]
        if area in self.universe.contents:
            self.set("location", area)
        self.universe.contents[area].contents[self.key] = self
        self.look_at(area)

    def go_home(self):
        """Relocate the element to its default location."""
        self.go_to(self.get("default_location"))
        self.echo_to_location(
            "You suddenly realize that " + self.get("name") + " is here."
        )

    def move_direction(self, direction):
        """Relocate the element in a specified direction."""
        motion = self.universe.contents["mudpy.movement.%s" % direction]
        enter_term = motion.get("enter_term")
        exit_term = motion.get("exit_term")
        self.echo_to_location("%s exits %s." % (self.get("name"), exit_term))
        self.send("You exit %s." % exit_term, add_prompt=False)
        self.go_to(
            self.universe.contents[
                self.get("location")].link_neighbor(direction)
        )
        self.echo_to_location("%s arrives from %s." % (
            self.get("name"), enter_term))

    def look_at(self, key):
        """Show an element to another element."""
        if self.owner:
            element = self.universe.contents[key]
            message = ""
            name = element.get("name")
            if name:
                message += "$(cyn)" + name + "$(nrm)$(eol)"
            description = element.get("description")
            if description:
                message += description + "$(eol)"
            portal_list = list(element.portals().keys())
            if portal_list:
                portal_list.sort()
                message += "$(cyn)[ Exits: " + ", ".join(
                    portal_list
                ) + " ]$(nrm)$(eol)"
            for element in self.universe.contents[
                self.get("location")
            ].contents.values():
                if element.get("is_actor") and element is not self:
                    message += "$(yel)" + element.get(
                        "name"
                    ) + " is here.$(nrm)$(eol)"
                elif element is not self:
                    message += "$(grn)" + element.get(
                        "impression"
                    ) + "$(nrm)$(eol)"
            self.send(message)

    def portals(self):
        """Map the portal directions for an area to neighbors."""
        portals = {}
        if re.match(r"""^area\.-?\d+,-?\d+,-?\d+$""", self.key):
            coordinates = [(int(x))
                           for x in self.key.split(".")[-1].split(",")]
            offsets = dict(
                (x,
                 self.universe.contents["mudpy.movement.%s" % x].get("vector")
                 ) for x in self.universe.directions)
            for portal in self.get("gridlinks"):
                adjacent = map(lambda c, o: c + o,
                               coordinates, offsets[portal])
                neighbor = "area." + ",".join(
                    [(str(x)) for x in adjacent]
                )
                if neighbor in self.universe.contents:
                    portals[portal] = neighbor
        for facet in self.facets():
            if facet.startswith("link_"):
                neighbor = self.get(facet)
                if neighbor in self.universe.contents:
                    portal = facet.split("_")[1]
                    portals[portal] = neighbor
        return portals

    def link_neighbor(self, direction):
        """Return the element linked in a given direction."""
        portals = self.portals()
        if direction in portals:
            return portals[direction]

    def echo_to_location(self, message):
        """Show a message to other elements in the current location."""
        for element in self.universe.contents[
            self.get("location")
        ].contents.values():
            if element is not self:
                element.send(message)


class Universe:

    """The universe."""

    def __init__(self, filename="", load=False):
        """Initialize the universe."""
        self.groups = {}
        self.contents = {}
        self.directions = set()
        self.loading = False
        self.loglines = []
        self.origins = {}
        self.reload_flag = False
        self.setup_loglines = []
        self.startdir = os.getcwd()
        self.terminate_flag = False
        self.userlist = []
        self.versions = None
        if not filename:
            possible_filenames = [
                "etc/mudpy.yaml",
                "/usr/local/mudpy/etc/mudpy.yaml",
                "/usr/local/etc/mudpy.yaml",
                "/etc/mudpy/mudpy.yaml",
                "/etc/mudpy.yaml"
            ]
            for filename in possible_filenames:
                if os.access(filename, os.R_OK):
                    break
        if not os.path.isabs(filename):
            filename = os.path.join(self.startdir, filename)
        self.filename = filename
        if load:
            # make sure to preserve any accumulated log entries during load
            self.setup_loglines += self.load()

    def load(self):
        """Load universe data from persistent storage."""

        # while loading, it's safe to update elements from read-only files
        self.loading = True

        # it's possible for this to enter before logging configuration is read
        pending_loglines = []

        # start populating the (re)files dict from the base config
        self.files = {}
        mudpy.data.Data(self.filename, self)

        # load default storage locations for groups
        if hasattr(self, "contents") and "mudpy.filing" in self.contents:
            self.origins.update(self.contents["mudpy.filing"].get(
                "groups", {}))

        # add some builtin groups we know we'll need
        for group in ("account", "actor", "internal"):
            self.add_group(group)

        # make a list of inactive avatars
        inactive_avatars = []
        for account in self.groups.get("account", {}).values():
            for avatar in account.get("avatars"):
                try:
                    inactive_avatars.append(self.contents[avatar])
                except KeyError:
                    pending_loglines.append((
                        'Missing avatar "%s", possible data corruption' %
                        avatar, 6))
        for user in self.userlist:
            if user.avatar in inactive_avatars:
                inactive_avatars.remove(user.avatar)

        # another pass to straighten out all the element contents
        for element in self.contents.values():
            element.update_location()
            element.clean_contents()

        # warn when debug mode has been engaged
        if self.debug_mode():
            pending_loglines.append((
                "WARNING: Unsafe debugging mode is enabled!", 6))

        # done loading, so disallow updating elements from read-only files
        self.loading = False

        return pending_loglines

    def new(self):
        """Create a new, empty Universe (the Big Bang)."""
        new_universe = Universe()
        for attribute in vars(self).keys():
            setattr(new_universe, attribute, getattr(self, attribute))
        new_universe.reload_flag = False
        del self
        return new_universe

    def save(self):
        """Save the universe to persistent storage."""
        for key in self.files:
            self.files[key].save()

    def initialize_server_socket(self):
        """Create and open the listening socket."""

        # need to know the local address and port number for the listener
        host = self.contents["mudpy.network"].get("host")
        port = self.contents["mudpy.network"].get("port")

        # if no host was specified, bind to the loopback address (preferring
        # ipv6)
        if not host:
            if socket.has_ipv6:
                host = "::1"
            else:
                host = "127.0.0.1"

        # figure out if this is ipv4 or v6
        family = socket.getaddrinfo(host, port)[0][0]
        if family is socket.AF_INET6 and not socket.has_ipv6:
            log("No support for IPv6 address %s (use IPv4 instead)." % host)

        # create a new stream-type socket object
        self.listening_socket = socket.socket(family, socket.SOCK_STREAM)

        # set the socket options to allow existing open ones to be
        # reused (fixes a bug where the server can't bind for a minute
        # when restarting on linux systems)
        self.listening_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1
        )

        # bind the socket to to our desired server ipa and port
        self.listening_socket.bind((host, port))

        # disable blocking so we can proceed whether or not we can
        # send/receive
        self.listening_socket.setblocking(0)

        # start listening on the socket
        self.listening_socket.listen(1)

        # note that we're now ready for user connections
        log("Listening for Telnet connections on %s port %s" % (
                host, str(port)))

    def get_time(self):
        """Convenience method to get the elapsed time counter."""
        try:
            return self.groups["internal"]["counters"].get("elapsed", 0)
        except KeyError:
            return 0

    def set_time(self, elapsed):
        """Convenience method to set the elapsed time counter."""
        try:
            self.groups["internal"]["counters"].set("elapsed", elapsed)
        except KeyError:
            # add an element for counters if it doesn't exist
            Element("internal.counters", universe)
            self.groups["internal"]["counters"].set("elapsed", elapsed)

    def add_group(self, group, fallback=None):
        """Set up group tracking/metadata."""
        if group not in self.origins:
            self.origins[group] = {}
        if not fallback:
            fallback = mudpy.data.find_file(
                    ".".join((group, "yaml")), universe=self)
        if "fallback" not in self.origins[group]:
            self.origins[group]["fallback"] = fallback
        flags = self.origins[group].get("flags", None)
        if fallback not in self.files:
            mudpy.data.Data(fallback, self, flags=flags)

    def debug_mode(self):
        """Boolean method to indicate whether unsafe debugging is enabled."""
        return self.groups["mudpy"]["limit"].get("debug", False)


class User:

    """This is a connected user."""

    def __init__(self):
        """Default values for the in-memory user variables."""
        self.account = None
        self.address = ""
        self.authenticated = False
        self.avatar = None
        self.choice = ""
        self.columns = 79
        self.connection = None
        self.error = ""
        self.input_queue = []
        self.last_address = ""
        self.last_input = universe.get_time()
        self.menu_choices = {}
        self.menu_seen = False
        self.negotiation_pause = 0
        self.output_queue = []
        self.partial_input = b""
        self.password_tries = 0
        self.rows = 23
        self.state = "telopt_negotiation"
        self.telopts = {}
        self.ttype = None
        self.universe = universe

    def quit(self):
        """Log, close the connection and remove."""
        if self.account:
            name = self.account.get("name", self)
        else:
            name = self
        log("Logging out %s" % name, 2)
        self.deactivate_avatar()
        self.connection.close()
        self.remove()

    def check_idle(self):
        """Warn or disconnect idle users as appropriate."""
        idletime = universe.get_time() - self.last_input
        linkdead_dict = universe.contents[
            "mudpy.timing.idle.disconnect"].facets()
        if self.state in linkdead_dict:
            linkdead_state = self.state
        else:
            linkdead_state = "default"
        if idletime > linkdead_dict[linkdead_state]:
            self.send(
                "$(eol)$(red)You've done nothing for far too long... goodbye!"
                + "$(nrm)$(eol)",
                flush=True,
                add_prompt=False
            )
            logline = "Disconnecting "
            if self.account and self.account.get("name"):
                logline += self.account.get("name")
            else:
                logline += "an unknown user"
            logline += (" after idling too long in the " + self.state
                        + " state.")
            log(logline, 2)
            self.state = "disconnecting"
            self.menu_seen = False
        idle_dict = universe.contents["mudpy.timing.idle.warn"].facets()
        if self.state in idle_dict:
            idle_state = self.state
        else:
            idle_state = "default"
        if idletime == idle_dict[idle_state]:
            self.send(
                "$(eol)$(red)If you continue to be unproductive, "
                + "you'll be shown the door...$(nrm)$(eol)"
            )

    def reload(self):
        """Save, load a new user and relocate the connection."""

        # copy old attributes
        attributes = self.__dict__

        # get out of the list
        self.remove()

        # get rid of the old user object
        del self

        # create a new user object
        new_user = User()

        # set everything equivalent
        new_user.__dict__ = attributes

        # the avatar needs a new owner
        if new_user.avatar:
            new_user.account = universe.contents[new_user.account.key]
            new_user.avatar = universe.contents[new_user.avatar.key]
            new_user.avatar.owner = new_user

        # add it to the list
        universe.userlist.append(new_user)

    def replace_old_connections(self):
        """Disconnect active users with the same name."""

        # the default return value
        return_value = False

        # iterate over each user in the list
        for old_user in universe.userlist:

            # the name is the same but it's not us
            if hasattr(
               old_user, "account"
               ) and old_user.account and old_user.account.get(
                "name"
            ) == self.account.get(
                "name"
            ) and old_user is not self:

                # make a note of it
                log(
                    "User " + self.account.get(
                        "name"
                    ) + " reconnected--closing old connection to "
                    + old_user.address + ".",
                    2
                )
                old_user.send(
                    "$(eol)$(red)New connection from " + self.address
                    + ". Terminating old connection...$(nrm)$(eol)",
                    flush=True,
                    add_prompt=False
                )

                # close the old connection
                old_user.connection.close()

                # replace the old connection with this one
                old_user.send(
                    "$(eol)$(red)Taking over old connection from "
                    + old_user.address + ".$(nrm)"
                )
                old_user.connection = self.connection
                old_user.last_address = old_user.address
                old_user.address = self.address
                old_user.telopts = self.telopts
                old_user.adjust_echoing()

                # take this one out of the list and delete
                self.remove()
                del self
                return_value = True
                break

        # true if an old connection was replaced, false if not
        return return_value

    def authenticate(self):
        """Flag the user as authenticated and disconnect duplicates."""
        if self.state != "authenticated":
            self.authenticated = True
            log("User %s authenticated for account %s." % (
                    self, self.account.subkey), 2)
            if ("mudpy.limit" in universe.contents and self.account.subkey in
                    universe.contents["mudpy.limit"].get("admins")):
                self.account.set("administrator", True)
                log("Account %s is an administrator." % (
                        self.account.subkey), 2)

    def show_menu(self):
        """Send the user their current menu."""
        if not self.menu_seen:
            self.menu_choices = get_menu_choices(self)
            self.send(
                get_menu(self.state, self.error, self.menu_choices),
                "",
                add_terminator=True
            )
            self.menu_seen = True
            self.error = False
            self.adjust_echoing()

    def prompt(self):
        """"Generate and return an input prompt."""

        # Start with the user's preference, if one was provided
        prompt = self.account.get("prompt")

        # If the user has not set a prompt, then immediately return the default
        # provided for the current state
        if not prompt:
            return get_menu_prompt(self.state)

        # Allow including the World clock state
        if "$_(time)" in prompt:
            prompt = prompt.replace(
                "$_(time)",
                str(universe.get_time()))

        # Append a single space for clear separation from user input
        if prompt[-1] != " ":
            prompt = "%s " % prompt

        # Return the cooked prompt
        return prompt

    def adjust_echoing(self):
        """Adjust echoing to match state menu requirements."""
        if mudpy.telnet.is_enabled(self, mudpy.telnet.TELOPT_ECHO,
                                   mudpy.telnet.US):
            if menu_echo_on(self.state):
                mudpy.telnet.disable(self, mudpy.telnet.TELOPT_ECHO,
                                     mudpy.telnet.US)
        elif not menu_echo_on(self.state):
            mudpy.telnet.enable(self, mudpy.telnet.TELOPT_ECHO,
                                mudpy.telnet.US)

    def remove(self):
        """Remove a user from the list of connected users."""
        log("Disconnecting account %s." % self, 0)
        universe.userlist.remove(self)

    def send(
        self,
        output,
        eol="$(eol)",
        raw=False,
        flush=False,
        add_prompt=True,
        just_prompt=False,
        add_terminator=False,
        prepend_padding=True
    ):
        """Send arbitrary text to a connected user."""

        # unless raw mode is on, clean it up all nice and pretty
        if not raw:

            # strip extra $(eol) off if present
            while output.startswith("$(eol)"):
                output = output[6:]
            while output.endswith("$(eol)"):
                output = output[:-6]
            extra_lines = output.find("$(eol)$(eol)$(eol)")
            while extra_lines > -1:
                output = output[:extra_lines] + output[extra_lines + 6:]
                extra_lines = output.find("$(eol)$(eol)$(eol)")

            # start with a newline, append the message, then end
            # with the optional eol string passed to this function
            # and the ansi escape to return to normal text
            if not just_prompt and prepend_padding:
                if (not self.output_queue or not
                        self.output_queue[-1].endswith(b"\r\n")):
                    output = "$(eol)" + output
                elif not self.output_queue[-1].endswith(
                    b"\r\n\x1b[0m\r\n"
                ) and not self.output_queue[-1].endswith(
                    b"\r\n\r\n"
                ):
                    output = "$(eol)" + output
            output += eol + chr(27) + "[0m"

            # tack on a prompt if active
            if self.state == "active":
                if not just_prompt:
                    output += "$(eol)"
                if add_prompt:
                    output += self.prompt()
                    mode = self.avatar.get("mode")
                    if mode:
                        output += "(" + mode + ") "

            # find and replace macros in the output
            output = replace_macros(self, output)

            # wrap the text at the client's width (min 40, 0 disables)
            if self.columns:
                if self.columns < 40:
                    wrap = 40
                else:
                    wrap = self.columns
                output = wrap_ansi_text(output, wrap)

            # if supported by the client, encode it utf-8
            if mudpy.telnet.is_enabled(self, mudpy.telnet.TELOPT_BINARY,
                                       mudpy.telnet.US):
                encoded_output = output.encode("utf-8")

            # otherwise just send ascii
            else:
                encoded_output = output.encode("ascii", "replace")

            # end with a terminator if requested
            if add_prompt or add_terminator:
                if mudpy.telnet.is_enabled(
                        self, mudpy.telnet.TELOPT_EOR, mudpy.telnet.US):
                    encoded_output += mudpy.telnet.telnet_proto(
                        mudpy.telnet.IAC, mudpy.telnet.EOR)
                elif not mudpy.telnet.is_enabled(
                        self, mudpy.telnet.TELOPT_SGA, mudpy.telnet.US):
                    encoded_output += mudpy.telnet.telnet_proto(
                        mudpy.telnet.IAC, mudpy.telnet.GA)

            # and tack it onto the queue
            self.output_queue.append(encoded_output)

            # if this is urgent, flush all pending output
            if flush:
                self.flush()

        # just dump raw bytes as requested
        else:
            self.output_queue.append(output)
            self.flush()

    def pulse(self):
        """All the things to do to the user per increment."""

        # if the world is terminating, disconnect
        if universe.terminate_flag:
            self.state = "disconnecting"
            self.menu_seen = False

        # check for an idle connection and act appropriately
        else:
            self.check_idle()

        # ask the client for their current terminal type (RFC 1091); it's None
        # if it's not been initialized, the empty string if it has but the
        # output was indeterminate, "UNKNOWN" if the client specified it has no
        # terminal types to supply
        if self.ttype is None:
            mudpy.telnet.request_ttype(self)

        # if output is paused, decrement the counter
        if self.state == "telopt_negotiation":
            if self.negotiation_pause:
                self.negotiation_pause -= 1
            else:
                self.state = "entering_account_name"

        # show the user a menu as needed
        elif not self.state == "active":
            self.show_menu()

        # flush any pending output in the queue
        self.flush()

        # disconnect users with the appropriate state
        if self.state == "disconnecting":
            self.quit()

        # check for input and add it to the queue
        self.enqueue_input()

        # there is input waiting in the queue
        if self.input_queue:
            handle_user_input(self)

    def flush(self):
        """Try to send the last item in the queue and remove it."""
        if self.output_queue:
            try:
                self.connection.send(self.output_queue[0])
            except (BrokenPipeError, ConnectionResetError):
                if self.account and self.account.get("name"):
                    account = self.account.get("name")
                else:
                    account = "an unknown user"
                self.state = "disconnecting"
                log("Disconnected while sending to %s." % account, 7)
            del self.output_queue[0]

    def enqueue_input(self):
        """Process and enqueue any new input."""

        # check for some input
        try:
            raw_input = self.connection.recv(1024)
        except OSError:
            raw_input = b""

        # we got something
        if raw_input:

            # tack this on to any previous partial
            self.partial_input += raw_input

            # reply to and remove any IAC negotiation codes
            mudpy.telnet.negotiate_telnet_options(self)

            # separate multiple input lines
            new_input_lines = self.partial_input.split(b"\r\0")
            if len(new_input_lines) == 1:
                new_input_lines = new_input_lines[0].split(b"\r\n")

            # if input doesn't end in a newline, replace the
            # held partial input with the last line of it
            if not (
                    self.partial_input.endswith(b"\r\0") or
                    self.partial_input.endswith(b"\r\n")):
                self.partial_input = new_input_lines.pop()

            # otherwise, chop off the extra null input and reset
            # the held partial input
            else:
                new_input_lines.pop()
                self.partial_input = b""

            # iterate over the remaining lines
            for line in new_input_lines:

                # strip off extra whitespace
                line = line.strip()

                # log non-printable characters remaining
                if not mudpy.telnet.is_enabled(
                        self, mudpy.telnet.TELOPT_BINARY, mudpy.telnet.HIM):
                    asciiline = bytes([x for x in line if 32 <= x <= 126])
                    if line != asciiline:
                        logline = "Non-ASCII characters from "
                        if self.account and self.account.get("name"):
                            logline += self.account.get("name") + ": "
                        else:
                            logline += "unknown user: "
                        logline += repr(line)
                        log(logline, 4)
                        line = asciiline

                try:
                    line = line.decode("utf-8")
                except UnicodeDecodeError:
                    logline = "Non-UTF-8 sequence from "
                    if self.account and self.account.get("name"):
                        logline += self.account.get("name") + ": "
                    else:
                        logline += "unknown user: "
                    logline += repr(line)
                    log(logline, 4)
                    return

                line = unicodedata.normalize("NFKC", line)

                # put on the end of the queue
                self.input_queue.append(line)

    def new_avatar(self):
        """Instantiate a new, unconfigured avatar for this user."""
        counter = 0
        while ("avatar_%s_%s" % (self.account.get("name"), counter)
                in universe.groups.get("actor", {}).keys()):
            counter += 1
        self.avatar = Element(
            "actor.avatar_%s_%s" % (self.account.get("name"), counter),
            universe)
        self.avatar.append("inherit", "archetype.avatar")
        self.account.append("avatars", self.avatar.key)
        log("Created new avatar %s for user %s." % (
                self.avatar.key, self.account.get("name")), 0)

    def delete_avatar(self, avatar):
        """Remove an avatar from the world and from the user's list."""
        if self.avatar is universe.contents[avatar]:
            self.avatar = None
        log("Deleting avatar %s for user %s." % (
                avatar, self.account.get("name")), 0)
        universe.contents[avatar].destroy()
        avatars = self.account.get("avatars")
        avatars.remove(avatar)
        self.account.set("avatars", avatars)

    def activate_avatar_by_index(self, index):
        """Enter the world with a particular indexed avatar."""
        self.avatar = universe.contents[
            self.account.get("avatars")[index]]
        self.avatar.owner = self
        self.state = "active"
        log("Activated avatar %s (%s)." % (
                self.avatar.get("name"), self.avatar.key), 0)
        self.avatar.go_home()

    def deactivate_avatar(self):
        """Have the active avatar leave the world."""
        if self.avatar:
            log("Deactivating avatar %s (%s) for user %s." % (
                    self.avatar.get("name"), self.avatar.key,
                    self.account.get("name")), 0)
            current = self.avatar.get("location")
            if current:
                self.avatar.set("default_location", current)
                self.avatar.echo_to_location(
                    "You suddenly wonder where " + self.avatar.get(
                        "name"
                    ) + " went."
                )
                del universe.contents[current].contents[self.avatar.key]
                self.avatar.remove_facet("location")
            self.avatar.owner = None
            self.avatar = None

    def destroy(self):
        """Destroy the user and associated avatars."""
        for avatar in self.account.get("avatars"):
            self.delete_avatar(avatar)
        log("Destroying account %s for user %s." % (
                self.account.get("name"), self), 0)
        self.account.destroy()

    def list_avatar_names(self):
        """List names of assigned avatars."""
        avatars = []
        for avatar in self.account.get("avatars"):
            try:
                avatars.append(universe.contents[avatar].get("name"))
            except KeyError:
                log('Missing avatar "%s", possible data corruption.' %
                    avatar, 6)
        return avatars

    def is_admin(self):
        """Boolean check whether user's account is an admin."""
        return self.account.get("administrator", False)


def broadcast(message, add_prompt=True):
    """Send a message to all connected users."""
    for each_user in universe.userlist:
        each_user.send("$(eol)" + message, add_prompt=add_prompt)


def log(message, level=0):
    """Log a message."""

    # a couple references we need
    if "mudpy.log" in universe.contents:
        file_name = universe.contents["mudpy.log"].get("file", "")
        max_log_lines = universe.contents["mudpy.log"].get("lines", 0)
        syslog_name = universe.contents["mudpy.log"].get("syslog", "")
    else:
        file_name = ""
        max_log_lines = 0
        syslog_name = ""
    timestamp = datetime.datetime.now().isoformat(' ')

    # turn the message into a list of nonempty lines
    lines = [x for x in [(x.rstrip()) for x in message.split("\n")] if x != ""]

    # send the timestamp and line to a file
    if file_name:
        if not os.path.isabs(file_name):
            file_name = os.path.join(universe.startdir, file_name)
        os.makedirs(os.path.dirname(file_name), exist_ok=True)
        file_descriptor = codecs.open(file_name, "a", "utf-8")
        for line in lines:
            file_descriptor.write(timestamp + " " + line + "\n")
        file_descriptor.flush()
        file_descriptor.close()

    # send the timestamp and line to standard output
    if ("mudpy.log" in universe.contents and
            universe.contents["mudpy.log"].get("stdout")):
        for line in lines:
            print(timestamp + " " + line)

    # send the line to the system log
    if syslog_name:
        syslog.openlog(
            syslog_name.encode("utf-8"),
            syslog.LOG_PID,
            syslog.LOG_INFO | syslog.LOG_DAEMON
        )
        for line in lines:
            syslog.syslog(line)
        syslog.closelog()

    # display to connected administrators
    for user in universe.userlist:
        if (
                user.state == "active"
                and user.is_admin()
                and user.account.get("loglevel", 0) <= level):
            # iterate over every line in the message
            full_message = ""
            for line in lines:
                full_message += (
                    "$(bld)$(red)" + timestamp + " "
                    + line.replace("$(", "$_(") + "$(nrm)$(eol)")
            user.send(full_message, flush=True)

    # add to the recent log list
    for line in lines:
        while 0 < len(universe.loglines) >= max_log_lines:
            del universe.loglines[0]
        universe.loglines.append((timestamp + " " + line, level))


def get_loglines(level, start, stop):
    """Return a specific range of loglines filtered by level."""

    # filter the log lines
    loglines = [x for x in universe.loglines if x[1] >= level]

    # we need these in several places
    total_count = str(len(universe.loglines))
    filtered_count = len(loglines)

    # don't proceed if there are no lines
    if filtered_count:

        # can't start before the beginning or at the end
        if start > filtered_count:
            start = filtered_count
        if start < 1:
            start = 1

        # can't stop before we start
        if stop > start:
            stop = start
        elif stop < 1:
            stop = 1

        # some preamble
        message = (
            "There are %s log lines in memory and %s at or above level %s. "
            "The matching lines from %s to %s are:$(eol)$(eol)" %
            (total_count, filtered_count, level, stop, start))

        # add the text from the selected lines
        if stop > 1:
            range_lines = loglines[-start:-(stop - 1)]
        else:
            range_lines = loglines[-start:]
        for line in range_lines:
            message += "   (%s) %s$(eol)" % (
                line[1], line[0].replace("$(", "$_("))

    # there were no lines
    else:
        message = "None of the %s lines in memory matches your request." % (
            total_count)

    # pass it back
    return message


def glyph_columns(character):
    """Convenience function to return the column width of a glyph."""
    if unicodedata.east_asian_width(character) in "FW":
        return 2
    else:
        return 1


def wrap_ansi_text(text, width):
    """Wrap text with arbitrary width while ignoring ANSI colors."""

    # the current position in the entire text string, including all
    # characters, printable or otherwise
    abs_pos = 0

    # the current text position relative to the beginning of the line,
    # ignoring color escape sequences
    rel_pos = 0

    # the absolute and relative positions of the most recent whitespace
    # character
    last_abs_whitespace = 0
    last_rel_whitespace = 0

    # whether the current character is part of a color escape sequence
    escape = False

    # normalize any potentially composited unicode before we count it
    text = unicodedata.normalize("NFKC", text)

    # iterate over each character from the beginning of the text
    for each_character in text:

        # the current character is the escape character
        if each_character == "\x1b" and not escape:
            escape = True
            rel_pos -= 1

        # the current character is within an escape sequence
        elif escape:
            rel_pos -= 1
            if each_character == "m":
                # the current character is m, which terminates the
                # escape sequence
                escape = False

        # the current character is a space
        elif each_character == " ":
            last_abs_whitespace = abs_pos
            last_rel_whitespace = rel_pos

        # the current character is a newline, so reset the relative
        # position too (start a new line)
        elif each_character == "\n":
            rel_pos = 0
            last_abs_whitespace = abs_pos
            last_rel_whitespace = rel_pos

        # the current character meets the requested maximum line width, so we
        # need to wrap unless the current word is wider than the terminal (in
        # which case we let it do the wrapping instead)
        if last_rel_whitespace != 0 and (rel_pos > width or (
                rel_pos > width - 1 and glyph_columns(each_character) == 2)):

            # insert an eol in place of the last space
            text = (text[:last_abs_whitespace] + "\r\n" +
                    text[last_abs_whitespace + 1:])

            # increase the absolute position because an eol is two
            # characters but the space it replaced was only one
            abs_pos += 1

            # now we're at the beginning of a new line, plus the
            # number of characters wrapped from the previous line
            rel_pos -= last_rel_whitespace
            last_rel_whitespace = 0

        # as long as the character is not a carriage return and the
        # other above conditions haven't been met, count it as a
        # printable character
        elif each_character != "\r":
            rel_pos += glyph_columns(each_character)
            if each_character in (" ", "\n"):
                last_abs_whitespace = abs_pos
                last_rel_whitespace = rel_pos

        # increase the absolute position for every character
        abs_pos += 1

    # return the newly-wrapped text
    return text


def weighted_choice(data):
    """Takes a dict weighted by value and returns a random key."""

    # this will hold our expanded list of keys from the data
    expanded = []

    # create the expanded list of keys
    for key in data.keys():
        for _count in range(data[key]):
            expanded.append(key)

    # return one at random
    # Allow the random.randrange() call in bandit since it's not used for
    # security/cryptographic purposes
    return random.choice(expanded)  # nosec


def random_name():
    """Returns a random character name."""

    # the vowels and consonants needed to create romaji syllables
    vowels = [
        "a",
        "i",
        "u",
        "e",
        "o"
    ]
    consonants = [
        "'",
        "k",
        "z",
        "s",
        "sh",
        "z",
        "j",
        "t",
        "ch",
        "ts",
        "d",
        "n",
        "h",
        "f",
        "m",
        "y",
        "r",
        "w"
    ]

    # this dict will hold our weighted list of syllables
    syllables = {}

    # generate the list with an even weighting
    for consonant in consonants:
        for vowel in vowels:
            syllables[consonant + vowel] = 1

    # we'll build the name into this string
    name = ""

    # create a name of random length from the syllables
    # Allow the random.randrange() call in bandit since it's not used for
    # security/cryptographic purposes
    for _syllable in range(random.randrange(2, 6)):  # nosec
        name += weighted_choice(syllables)

    # strip any leading quotemark, capitalize and return the name
    return name.strip("'").capitalize()


def replace_macros(user, text, is_input=False):
    """Replaces macros in text output."""

    # third person pronouns
    pronouns = {
        "female": {"obj": "her", "pos": "hers", "sub": "she"},
        "male": {"obj": "him", "pos": "his", "sub": "he"},
        "neuter": {"obj": "it", "pos": "its", "sub": "it"}
    }

    # a dict of replacement macros
    macros = {
        "eol": "\r\n",
        "bld": chr(27) + "[1m",
        "nrm": chr(27) + "[0m",
        "blk": chr(27) + "[30m",
        "blu": chr(27) + "[34m",
        "cyn": chr(27) + "[36m",
        "grn": chr(27) + "[32m",
        "mgt": chr(27) + "[35m",
        "red": chr(27) + "[31m",
        "yel": chr(27) + "[33m",
    }

    # add dynamic macros where possible
    if user.account:
        account_name = user.account.get("name")
        if account_name:
            macros["account"] = account_name
    if user.avatar:
        avatar_gender = user.avatar.get("gender")
        if avatar_gender:
            macros["tpop"] = pronouns[avatar_gender]["obj"]
            macros["tppp"] = pronouns[avatar_gender]["pos"]
            macros["tpsp"] = pronouns[avatar_gender]["sub"]

    # loop until broken
    while True:

        # find and replace per the macros dict
        macro_start = text.find("$(")
        if macro_start == -1:
            break
        macro_end = text.find(")", macro_start) + 1
        macro = text[macro_start + 2:macro_end - 1]
        if macro in macros.keys():
            replacement = macros[macro]

        # this is how we handle local file inclusion (dangerous!)
        elif macro.startswith("inc:"):
            incfile = mudpy.data.find_file(macro[4:], universe=universe)
            if os.path.exists(incfile):
                replacement = ""
                with codecs.open(incfile, "r", "utf-8") as incfd:
                    for line in incfd:
                        if line.endswith("\n") and not line.endswith("\r\n"):
                            line = line.replace("\n", "\r\n")
                        replacement += line
                    # lose the trailing eol
                    replacement = replacement[:-2]
            else:
                replacement = ""
                log("Couldn't read included " + incfile + " file.", 7)

        # if we get here, log and replace it with null
        else:
            replacement = ""
            if not is_input:
                log("Unexpected replacement macro " +
                    macro + " encountered.", 6)

        # and now we act on the replacement
        text = text.replace("$(" + macro + ")", replacement)

    # replace the look-like-a-macro sequence
    text = text.replace("$_(", "$(")

    return text


def escape_macros(value):
    """Escapes replacement macros in text."""
    if type(value) is str:
        return value.replace("$(", "$_(")
    else:
        return value


def first_word(text, separator=" "):
    """Returns a tuple of the first word and the rest."""
    if text:
        if text.find(separator) > 0:
            return text.split(separator, 1)
        else:
            return text, ""
    else:
        return "", ""


def on_pulse():
    """The things which should happen on each pulse, aside from reloads."""

    # increase the elapsed increment counter
    universe.set_time(universe.get_time() + 1)

    # update the log every now and then
    if not universe.groups["internal"]["counters"].get("mark"):
        log(str(len(universe.userlist)) + " connection(s)")
        universe.groups["internal"]["counters"].set(
            "mark", universe.contents["mudpy.timing"].get("status")
        )
    else:
        universe.groups["internal"]["counters"].set(
            "mark", universe.groups["internal"]["counters"].get(
                "mark"
            ) - 1
        )

    # periodically save everything
    if not universe.groups["internal"]["counters"].get("save"):
        universe.save()
        universe.groups["internal"]["counters"].set(
            "save", universe.contents["mudpy.timing"].get("save")
        )
    else:
        universe.groups["internal"]["counters"].set(
            "save", universe.groups["internal"]["counters"].get(
                "save"
            ) - 1
        )

    # open the listening socket if it hasn't been already
    if not hasattr(universe, "listening_socket"):
        universe.initialize_server_socket()

    # assign a user if a new connection is waiting
    user = check_for_connection(universe.listening_socket)
    if user:
        universe.userlist.append(user)

    # iterate over the connected users
    for user in universe.userlist:
        user.pulse()

    # pause for a configurable amount of time (decimal seconds)
    time.sleep(universe.contents["mudpy.timing"].get("increment"))


def reload_data():
    """Reload all relevant objects."""
    universe.save()
    old_userlist = universe.userlist[:]
    old_loglines = universe.loglines[:]
    for element in list(universe.contents.values()):
        element.destroy()
    pending_loglines = universe.load()
    new_loglines = universe.loglines[:]
    universe.loglines = old_loglines + new_loglines + pending_loglines
    for user in old_userlist:
        user.reload()


def check_for_connection(listening_socket):
    """Check for a waiting connection and return a new user object."""

    # try to accept a new connection
    try:
        connection, address = listening_socket.accept()
    except BlockingIOError:
        return None

    # note that we got one
    log("New connection from %s." % address[0], 2)

    # disable blocking so we can proceed whether or not we can send/receive
    connection.setblocking(0)

    # create a new user object
    user = User()
    log("Instantiated %s for %s." % (user, address[0]), 0)

    # associate this connection with it
    user.connection = connection

    # set the user's ipa from the connection's ipa
    user.address = address[0]

    # let the client know we WILL EOR (RFC 885)
    mudpy.telnet.enable(user, mudpy.telnet.TELOPT_EOR, mudpy.telnet.US)
    user.negotiation_pause = 2

    # return the new user object
    return user


def find_command(command_name):
    """Try to find a command by name or abbreviation."""

    # lowercase the command
    command_name = command_name.lower()

    command = None
    if command_name in universe.groups["command"]:
        # the command matches a command word for which we have data
        command = universe.groups["command"][command_name]
    else:
        for candidate in sorted(universe.groups["command"]):
            if candidate.startswith(command_name) and not universe.groups[
                    "command"][candidate].is_restricted():
                # the command matches the start of a command word and is not
                # restricted to administrators
                command = universe.groups["command"][candidate]
                break
    return command


def get_menu(state, error=None, choices=None):
    """Show the correct menu text to a user."""

    # make sure we don't reuse a mutable sequence by default
    if choices is None:
        choices = {}

    # get the description or error text
    message = get_menu_description(state, error)

    # get menu choices for the current state
    message += get_formatted_menu_choices(state, choices)

    # try to get a prompt, if it was defined
    message += get_menu_prompt(state)

    # throw in the default choice, if it exists
    message += get_formatted_default_menu_choice(state)

    # display a message indicating if echo is off
    message += get_echo_message(state)

    # return the assembly of various strings defined above
    return message


def menu_echo_on(state):
    """True if echo is on, false if it is off."""
    return universe.groups["menu"][state].get("echo", True)


def get_echo_message(state):
    """Return a message indicating that echo is off."""
    if menu_echo_on(state):
        return ""
    else:
        return "(won't echo) "


def get_default_menu_choice(state):
    """Return the default choice for a menu."""
    return universe.groups["menu"][state].get("default")


def get_formatted_default_menu_choice(state):
    """Default menu choice foratted for inclusion in a prompt string."""
    default_choice = get_default_menu_choice(state)
    if default_choice:
        return "[$(red)" + default_choice + "$(nrm)] "
    else:
        return ""


def get_menu_description(state, error):
    """Get the description or error text."""

    # an error condition was raised by the handler
    if error:

        # try to get an error message matching the condition
        # and current state
        description = universe.groups[
            "menu"][state].get("error_" + error)
        if not description:
            description = "That is not a valid choice..."
        description = "$(red)" + description + "$(nrm)"

    # there was no error condition
    else:

        # try to get a menu description for the current state
        description = universe.groups["menu"][state].get("description")

    # return the description or error message
    if description:
        description += "$(eol)$(eol)"
    return description


def get_menu_prompt(state):
    """Try to get a prompt, if it was defined."""
    prompt = universe.groups["menu"][state].get("prompt")
    if prompt:
        prompt += " "
    return prompt


def get_menu_choices(user):
    """Return a dict of choice:meaning."""
    state = universe.groups["menu"][user.state]
    create_choices = state.get("create")
    if create_choices:
        choices = call_hook_function(create_choices, (user,))
    else:
        choices = {}
    ignores = []
    options = {}
    creates = {}
    for facet in state.facets():
        if facet.startswith("demand_") and not call_hook_function(
                universe.groups["menu"][user.state].get(facet), (user,)):
            ignores.append(facet.split("_", 2)[1])
        elif facet.startswith("create_"):
            creates[facet] = facet.split("_", 2)[1]
        elif facet.startswith("choice_"):
            options[facet] = facet.split("_", 2)[1]
    for facet in creates.keys():
        if not creates[facet] in ignores:
            choices[creates[facet]] = call_hook_function(
                state.get(facet), (user,))
    for facet in options.keys():
        if not options[facet] in ignores:
            choices[options[facet]] = state.get(facet)
    return choices


def get_formatted_menu_choices(state, choices):
    """Returns a formatted string of menu choices."""
    choice_output = ""
    choice_keys = list(choices.keys())
    choice_keys.sort()
    for choice in choice_keys:
        choice_output += "   [$(red)" + choice + "$(nrm)]  " + choices[
            choice
        ] + "$(eol)"
    if choice_output:
        choice_output += "$(eol)"
    return choice_output


def get_menu_branches(state):
    """Return a dict of choice:branch."""
    branches = {}
    for facet in universe.groups["menu"][state].facets():
        if facet.startswith("branch_"):
            branches[
                facet.split("_", 2)[1]
            ] = universe.groups["menu"][state].get(facet)
    return branches


def get_default_branch(state):
    """Return the default branch."""
    return universe.groups["menu"][state].get("branch")


def get_choice_branch(user):
    """Returns the new state matching the given choice."""
    branches = get_menu_branches(user.state)
    if user.choice in branches.keys():
        return branches[user.choice]
    elif user.choice in user.menu_choices.keys():
        return get_default_branch(user.state)
    else:
        return ""


def get_menu_actions(state):
    """Return a dict of choice:branch."""
    actions = {}
    for facet in universe.groups["menu"][state].facets():
        if facet.startswith("action_"):
            actions[
                facet.split("_", 2)[1]
            ] = universe.groups["menu"][state].get(facet)
    return actions


def get_default_action(state):
    """Return the default action."""
    return universe.groups["menu"][state].get("action")


def get_choice_action(user):
    """Run any indicated script for the given choice."""
    actions = get_menu_actions(user.state)
    if user.choice in actions.keys():
        return actions[user.choice]
    elif user.choice in user.menu_choices.keys():
        return get_default_action(user.state)
    else:
        return ""


def call_hook_function(fname, arglist):
    """Safely execute named function with supplied arguments, return result."""

    # all functions relative to mudpy package
    function = mudpy

    for component in fname.split("."):
        try:
            function = getattr(function, component)
        except AttributeError:
            log('Could not find mudpy.%s() for arguments "%s"'
                % (fname, arglist), 7)
            function = None
            break
    if function:
        try:
            return function(*arglist)
        except Exception:
            log('Calling mudpy.%s(%s) raised an exception...\n%s'
                % (fname, (*arglist,), traceback.format_exc()), 7)


def handle_user_input(user):
    """The main handler, branches to a state-specific handler."""

    # if the user's client echo is off, send a blank line for aesthetics
    if mudpy.telnet.is_enabled(user, mudpy.telnet.TELOPT_ECHO,
                               mudpy.telnet.US):
        user.send("", add_prompt=False, prepend_padding=False)

    # check to make sure the state is expected, then call that handler
    try:
        globals()["handler_" + user.state](user)
    except KeyError:
        generic_menu_handler(user)

    # since we got input, flag that the menu/prompt needs to be redisplayed
    user.menu_seen = False

    # update the last_input timestamp while we're at it
    user.last_input = universe.get_time()


def generic_menu_handler(user):
    """A generic menu choice handler."""

    # get a lower-case representation of the next line of input
    if user.input_queue:
        user.choice = user.input_queue.pop(0)
        if user.choice:
            user.choice = user.choice.lower()
    else:
        user.choice = ""
    if not user.choice:
        user.choice = get_default_menu_choice(user.state)
    if user.choice in user.menu_choices:
        action = get_choice_action(user)
        if action:
            call_hook_function(action, (user,))
        new_state = get_choice_branch(user)
        if new_state:
            user.state = new_state
    else:
        user.error = "default"


def handler_entering_account_name(user):
    """Handle the login account name."""

    # get the next waiting line of input
    input_data = user.input_queue.pop(0)

    # did the user enter anything?
    if input_data:

        # keep only the first word and convert to lower-case
        name = input_data.lower()

        # fail if there are non-alphanumeric characters
        if name != "".join(filter(
                lambda x: x >= "0" and x <= "9" or x >= "a" and x <= "z",
                name)):
            user.error = "bad_name"

        # if that account exists, time to request a password
        elif name in universe.groups.get("account", {}):
            user.account = universe.groups["account"][name]
            user.state = "checking_password"

        # otherwise, this could be a brand new user
        else:
            user.account = Element("account.%s" % name, universe)
            user.account.set("name", name)
            log("New user: " + name, 2)
            user.state = "checking_new_account_name"

    # if the user entered nothing for a name, then buhbye
    else:
        user.state = "disconnecting"


def handler_checking_password(user):
    """Handle the login account password."""

    # get the next waiting line of input
    input_data = user.input_queue.pop(0)

    if "mudpy.limit" in universe.contents:
        max_password_tries = universe.contents["mudpy.limit"].get(
            "password_tries", 3)
    else:
        max_password_tries = 3

    # does the hashed input equal the stored hash?
    if mudpy.password.verify(input_data, user.account.get("passhash")):

        # if so, set the username and load from cold storage
        if not user.replace_old_connections():
            user.authenticate()
            user.state = "main_utility"

    # if at first your hashes don't match, try, try again
    elif user.password_tries < max_password_tries - 1:
        user.password_tries += 1
        user.error = "incorrect"

    # we've exceeded the maximum number of password failures, so disconnect
    else:
        user.send(
            "$(eol)$(red)Too many failed password attempts...$(nrm)$(eol)"
        )
        user.state = "disconnecting"


def handler_entering_new_password(user):
    """Handle a new password entry."""

    # get the next waiting line of input
    input_data = user.input_queue.pop(0)

    if "mudpy.limit" in universe.contents:
        max_password_tries = universe.contents["mudpy.limit"].get(
            "password_tries", 3)
    else:
        max_password_tries = 3

    # make sure the password is strong--at least one upper, one lower and
    # one digit, seven or more characters in length
    if len(input_data) > 6 and len(
       list(filter(lambda x: x >= "0" and x <= "9", input_data))
       ) and len(
        list(filter(lambda x: x >= "A" and x <= "Z", input_data))
    ) and len(
        list(filter(lambda x: x >= "a" and x <= "z", input_data))
    ):

        # hash and store it, then move on to verification
        user.account.set("passhash", mudpy.password.create(input_data))
        user.state = "verifying_new_password"

    # the password was weak, try again if you haven't tried too many times
    elif user.password_tries < max_password_tries - 1:
        user.password_tries += 1
        user.error = "weak"

    # too many tries, so adios
    else:
        user.send(
            "$(eol)$(red)Too many failed password attempts...$(nrm)$(eol)"
        )
        user.account.destroy()
        user.state = "disconnecting"


def handler_verifying_new_password(user):
    """Handle the re-entered new password for verification."""

    # get the next waiting line of input
    input_data = user.input_queue.pop(0)

    if "mudpy.limit" in universe.contents:
        max_password_tries = universe.contents["mudpy.limit"].get(
            "password_tries", 3)
    else:
        max_password_tries = 3

    # hash the input and match it to storage
    if mudpy.password.verify(input_data, user.account.get("passhash")):
        user.authenticate()

        # the hashes matched, so go active
        if not user.replace_old_connections():
            user.state = "main_utility"

    # go back to entering the new password as long as you haven't tried
    # too many times
    elif user.password_tries < max_password_tries - 1:
        user.password_tries += 1
        user.error = "differs"
        user.state = "entering_new_password"

    # otherwise, sayonara
    else:
        user.send(
            "$(eol)$(red)Too many failed password attempts...$(nrm)$(eol)"
        )
        user.account.destroy()
        user.state = "disconnecting"


def handler_active(user):
    """Handle input for active users."""

    # get the next waiting line of input
    input_data = user.input_queue.pop(0)

    # is there input?
    if input_data:

        # split out the command and parameters
        actor = user.avatar
        mode = actor.get("mode")
        if mode and input_data.startswith("!"):
            command_name, parameters = first_word(input_data[1:])
        elif mode == "chat":
            command_name = "say"
            parameters = input_data
        else:
            command_name, parameters = first_word(input_data)

        # expand to an actual command
        command = find_command(command_name)

        # if it's allowed, do it
        result = None
        if actor.can_run(command):
            action_fname = command.get("action", command.key)
            if action_fname:
                result = call_hook_function(action_fname, (actor, parameters))

        # if the command was not run, give an error
        if not result:
            mudpy.command.error(actor, input_data)

    # if no input, just idle back with a prompt
    else:
        user.send("", just_prompt=True)


def daemonize(universe):
    """Fork and disassociate from everything."""

    # only if this is what we're configured to do
    if "mudpy.process" in universe.contents and universe.contents[
            "mudpy.process"].get("daemon"):

        # log before we start forking around, so the terminal gets the message
        log("Disassociating from the controlling terminal.")

        # fork off and die, so we free up the controlling terminal
        if os.fork():
            os._exit(0)

        # switch to a new process group
        os.setsid()

        # fork some more, this time to free us from the old process group
        if os.fork():
            os._exit(0)

        # reset the working directory so we don't needlessly tie up mounts
        os.chdir("/")

        # clear the file creation mask so we can bend it to our will later
        os.umask(0)

        # redirect stdin/stdout/stderr and close off their former descriptors
        for stdpipe in range(3):
            os.close(stdpipe)
        sys.stdin = codecs.open("/dev/null", "r", "utf-8")
        sys.__stdin__ = codecs.open("/dev/null", "r", "utf-8")
        sys.stdout = codecs.open("/dev/null", "w", "utf-8")
        sys.stderr = codecs.open("/dev/null", "w", "utf-8")
        sys.__stdout__ = codecs.open("/dev/null", "w", "utf-8")
        sys.__stderr__ = codecs.open("/dev/null", "w", "utf-8")


def create_pidfile(universe):
    """Write a file containing the current process ID."""
    pid = str(os.getpid())
    log("Process ID: " + pid)
    if "mudpy.process" in universe.contents:
        file_name = universe.contents["mudpy.process"].get("pidfile", "")
    else:
        file_name = ""
    if file_name:
        if not os.path.isabs(file_name):
            file_name = os.path.join(universe.startdir, file_name)
        os.makedirs(os.path.dirname(file_name), exist_ok=True)
        file_descriptor = codecs.open(file_name, "w", "utf-8")
        file_descriptor.write(pid + "\n")
        file_descriptor.flush()
        file_descriptor.close()


def remove_pidfile(universe):
    """Remove the file containing the current process ID."""
    if "mudpy.process" in universe.contents:
        file_name = universe.contents["mudpy.process"].get("pidfile", "")
    else:
        file_name = ""
    if file_name:
        if not os.path.isabs(file_name):
            file_name = os.path.join(universe.startdir, file_name)
        if os.access(file_name, os.W_OK):
            os.remove(file_name)


def excepthook(excepttype, value, tracebackdata):
    """Handle uncaught exceptions."""

    # assemble the list of errors into a single string
    message = "".join(
        traceback.format_exception(excepttype, value, tracebackdata)
    )

    # try to log it, if possible
    try:
        log(message, 9)
    except Exception as e:
        # try to write it to stderr, if possible
        sys.stderr.write(message + "\nException while logging...\n%s" % e)


def sighook(what, where):
    """Handle external signals."""

    # a generic message
    message = "Caught signal: "

    # for a hangup signal
    if what == signal.SIGHUP:
        message += "hangup (reloading)"
        universe.reload_flag = True

    # for a terminate signal
    elif what == signal.SIGTERM:
        message += "terminate (halting)"
        universe.terminate_flag = True

    # catchall for unexpected signals
    else:
        message += str(what) + " (unhandled)"

    # log what happened
    log(message, 8)


def override_excepthook():
    """Redefine sys.excepthook with our own."""
    sys.excepthook = excepthook


def assign_sighook():
    """Assign a customized handler for some signals."""
    signal.signal(signal.SIGHUP, sighook)
    signal.signal(signal.SIGTERM, sighook)


def setup():
    """This contains functions to be performed when starting the engine."""

    # see if a configuration file was specified
    if len(sys.argv) > 1:
        conffile = sys.argv[1]
    else:
        conffile = ""

    # the big bang
    global universe
    universe = Universe(conffile, True)

    # report any loglines which accumulated during setup
    for logline in universe.setup_loglines:
        log(*logline)
    universe.setup_loglines = []

    # fork and disassociate
    daemonize(universe)

    # override the default exception handler so we get logging first thing
    override_excepthook()

    # set up custom signal handlers
    assign_sighook()

    # make the pidfile
    create_pidfile(universe)

    # load and store diagnostic info
    universe.versions = mudpy.version.Versions("mudpy")

    # log startup diagnostic messages
    log("On %s at %s" % (universe.versions.python_version, sys.executable), 1)
    log("Import path: %s" % ", ".join(sys.path), 1)
    log("Installed dependencies: %s" % universe.versions.dependencies_text, 1)
    log("Other python packages: %s" % universe.versions.environment_text, 1)
    log("Running version: %s" % universe.versions.version, 1)
    log("Initial directory: %s" % universe.startdir, 1)
    log("Command line: %s" % " ".join(sys.argv), 1)

    # pass the initialized universe back
    return universe


def finish():
    """These are functions performed when shutting down the engine."""

    # the loop has terminated, so save persistent data
    universe.save()

    # log a final message
    log("Shutting down now.")

    # get rid of the pidfile
    remove_pidfile(universe)



"""Telnet functions and constants for the mudpy engine."""

# Copyright (c) 2004-2020 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import mudpy

# telnet options (from bsd's arpa/telnet.h since telnetlib's are ambiguous)
TELOPT_BINARY = 0  # transmit 8-bit data by the receiver (rfc 856)
TELOPT_ECHO = 1  # echo received data back to the sender (rfc 857)
TELOPT_SGA = 3  # suppress transmission of the go ahead character (rfc 858)
TELOPT_TTYPE = 24  # exchange terminal type information (rfc 1091)
TELOPT_EOR = 25  # transmit end-of-record after transmitting data (rfc 885)
TELOPT_NAWS = 31  # negotiate about window size (rfc 1073)
TELOPT_LINEMODE = 34  # cooked line-by-line input mode (rfc 1184)
option_names = {
    TELOPT_BINARY: "8-bit binary",
    TELOPT_ECHO: "echo",
    TELOPT_SGA: "suppress go-ahead",
    TELOPT_TTYPE: "terminal type",
    TELOPT_EOR: "end of record",
    TELOPT_NAWS: "negotiate about window size",
    TELOPT_LINEMODE: "line mode",
}

supported = (
    TELOPT_BINARY,
    TELOPT_ECHO,
    TELOPT_SGA,
    TELOPT_TTYPE,
    TELOPT_EOR,
    TELOPT_NAWS,
    TELOPT_LINEMODE
)

# telnet commands
EOR = 239  # end-of-record signal (rfc 885)
SE = 240  # end of subnegotiation parameters (rfc 854)
GA = 249  # go ahead signal (rfc 854)
SB = 250  # what follows is subnegotiation of the indicated option (rfc 854)
WILL = 251  # desire or confirmation of performing an option (rfc 854)
WONT = 252  # refusal or confirmation of performing an option (rfc 854)
DO = 253  # request or confirm performing an option (rfc 854)
DONT = 254  # demand or confirm no longer performing an option (rfc 854)
IAC = 255  # interpret as command escape character (rfc 854)
command_names = {
    EOR: "end of record",
    SE: "subnegotiation end",
    GA: "go ahead",
    SB: "subnegotiation begin",
    WILL: "will",
    WONT: "won't",
    DO: "do",
    DONT: "don't",
    IAC: "interpret as command",
}

# RFC 1143 option negotiation states
NO = 0  # option is disabled
YES = 1  # option is enabled
WANTNO = 2  # demanded disabling option
WANTYES = 3  # requested enabling option
WANTNO_OPPOSITE = 4  # demanded disabling option but queued an enable after it
WANTYES_OPPOSITE = 5  # requested enabling option but queued a disable after it
state_names = {
    NO: "disabled",
    YES: "enabled",
    WANTNO: "demand disabling",
    WANTYES: "request enabling",
    WANTNO_OPPOSITE: "want no queued opposite",
    WANTYES_OPPOSITE: "want yes queued opposite",
}

# RFC 1143 option negotiation parties
HIM = 0
US = 1
party_names = {
    HIM: "him",
    US: "us",
}

# RFC 1091 commands
IS = 0
SEND = 1
ttype_command_names = {
    IS: "is",
    SEND: "send",
}


def log(message, user):
    """Log debugging info for Telnet client/server interactions."""
    if user.account:
        client = user.account.get("name", user)
    else:
        client = user
    mudpy.misc.log('[telnet] %s %s.' % (message, client), 0)


def telnet_proto(*arguments):
    """Return a concatenated series of Telnet protocol commands."""
    return bytes((arguments))


def translate_action(*command):
    """Convert a Telnet command sequence into text suitable for logging."""
    try:
        command_name = command_names[command[0]]
    except KeyError:
        # This should never happen since we filter unknown commands from
        # the input queue, but added here for completeness since logging
        # should never crash the process
        command_name = str(command[0])
    try:
        option_name = option_names[command[1]]
    except KeyError:
        # This can happen for any of the myriad of Telnet options missing
        # from the option_names dict
        option_name = str(command[1])
    return "%s %s" % (command_name, option_name)


def send_command(user, *command):
    """Sends a Telnet command string to the specified user's socket."""
    user.send(telnet_proto(IAC, *command), raw=True)
    log('Sent "%s" to' % translate_action(*command), user)


def is_enabled(user, telopt, party, state=YES):
    """Indicates whether a specified Telnet option is enabled."""
    if (telopt, party) in user.telopts and user.telopts[
       (telopt, party)
       ] is state:
        return True
    else:
        return False


def enable(user, telopt, party):
    """Negotiates enabling a Telnet option for the indicated user's socket."""
    if party is HIM:
        txpos = DO
    else:
        txpos = WILL
    if not (telopt, party) in user.telopts or user.telopts[
       (telopt, party)
       ] is NO:
        user.telopts[(telopt, party)] = WANTYES
        send_command(user, txpos, telopt)
    elif user.telopts[(telopt, party)] is WANTNO:
        user.telopts[(telopt, party)] = WANTNO_OPPOSITE
    elif user.telopts[(telopt, party)] is WANTYES_OPPOSITE:
        user.telopts[(telopt, party)] = WANTYES


def disable(user, telopt, party):
    """Negotiates disabling a Telnet option for the user's socket."""
    if party is HIM:
        txneg = DONT
    else:
        txneg = WONT
    if not (telopt, party) in user.telopts:
        user.telopts[(telopt, party)] = NO
    elif user.telopts[(telopt, party)] is YES:
        user.telopts[(telopt, party)] = WANTNO
        send_command(user, txneg, telopt)
    elif user.telopts[(telopt, party)] is WANTYES:
        user.telopts[(telopt, party)] = WANTYES_OPPOSITE
    elif user.telopts[(telopt, party)] is WANTNO_OPPOSITE:
        user.telopts[(telopt, party)] = WANTNO


def request_ttype(user):
    """Clear and request the terminal type."""

    # only actually request if the corresponding telopt is enabled
    if is_enabled(user, TELOPT_TTYPE, HIM):
        # set to the empty string to indicate it's been requested
        user.ttype = ""
        user.send(telnet_proto(IAC, SB, TELOPT_TTYPE, SEND, IAC, SE), raw=True)
        log('Sent terminal type request to', user)


def negotiate_telnet_options(user):
    """Reply to and remove telnet negotiation options from partial_input."""

    # make a local copy to play with
    text = user.partial_input

    # start at the beginning of the input
    position = 0

    # as long as we haven't checked it all
    len_text = len(text)
    while position < len_text:

        # jump to the first IAC you find
        position = text.find(telnet_proto(IAC), position)

        # if there wasn't an IAC in the input or it's at the end, we're done
        if position < 0 or position >= len_text - 1:
            break

        # the byte following the IAC is our command
        command = text[position+1]

        # replace a double (literal) IAC if there's a CR+NUL or CR+LF later
        if command is IAC:
            if (
                    text.find(b"\r\0", position) > 0 or
                    text.find(b"\r\n", position) > 0):
                position += 1
                text = text[:position] + text[position + 1:]
                log('Escaped IAC from', user)
            else:
                position += 2

        # implement an RFC 1143 option negotiation queue here
        elif len_text > position + 2 and WILL <= command <= DONT:
            telopt = text[position+2]
            log('Received "%s" from' % translate_action(command, telopt), user)
            if telopt in supported:
                if command <= WONT:
                    party = HIM
                    rxpos = WILL
                    txpos = DO
                    txneg = DONT
                else:
                    party = US
                    rxpos = DO
                    txpos = WILL
                    txneg = WONT
                if (telopt, party) not in user.telopts:
                    user.telopts[(telopt, party)] = NO
                if command is rxpos:
                    if user.telopts[(telopt, party)] is NO:
                        user.telopts[(telopt, party)] = YES
                        send_command(user, txpos, telopt)
                    elif user.telopts[(telopt, party)] is WANTNO:
                        user.telopts[(telopt, party)] = NO
                    elif user.telopts[(telopt, party)] is WANTNO_OPPOSITE:
                        user.telopts[(telopt, party)] = YES
                    elif user.telopts[(telopt, party)] is WANTYES_OPPOSITE:
                        user.telopts[(telopt, party)] = WANTNO
                        send_command(user, txneg, telopt)
                    else:
                        user.telopts[(telopt, party)] = YES
                else:
                    if user.telopts[(telopt, party)] is YES:
                        user.telopts[(telopt, party)] = NO
                        send_command(user, txneg, telopt)
                    elif user.telopts[(telopt, party)] is WANTNO_OPPOSITE:
                        user.telopts[(telopt, party)] = WANTYES
                        send_command(user, txpos, telopt)
                    else:
                        user.telopts[(telopt, party)] = NO
            elif command is WILL:
                send_command(user, DONT, telopt)
            else:
                send_command(user, WONT, telopt)
            text = text[:position] + text[position + 3:]

        # subnegotiation options
        elif len_text > position + 4 and command is SB:
            telopt = text[position + 2]
            end_subnegotiation = text.find(telnet_proto(IAC, SE), position)
            if end_subnegotiation > 0:
                if telopt is TELOPT_NAWS:
                    user.columns = (
                        text[position + 3] * 256 + text[position + 4])
                    user.rows = (
                        text[position + 5] * 256 + text[position + 6])
                elif telopt is TELOPT_TTYPE and text[position + 3] is IS:
                    user.ttype = (
                        text[position + 4:end_subnegotiation]).decode("ascii")
                text = text[:position] + text[end_subnegotiation + 2:]
            else:
                position += 1

        # otherwise, strip out a two-byte IAC command
        elif len_text > position + 2:
            log("Ignored unknown command %s from" % command, user)
            text = text[:position] + text[position + 2:]

        # and this means we got the beginning of an IAC
        else:
            position += 1

    # replace the input with our cleaned-up text
    user.partial_input = text



# Copyright (c) 2016-2019 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import setuptools

setuptools.setup(setup_requires=['pbr>=5'], pbr=True)



# Copyright (c) 2018-2021 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import collections
import os
import sys

sys.path.insert(0, os.path.abspath('../..'))

project = 'mudpy'

add_function_parentheses = True
add_module_names = True
copyright = '2004-2021, mudpy authors'
extensions = ['sphinx.ext.autodoc']
html_favicon = '_static/logo.svg'
html_logo = '_static/logo.svg'
html_sidebars = {'**': [
    'about.html',
    'donate.html',
    'navigation.html',
    'relations.html',
    'searchbox.html',
]}
html_theme_options = {
    'description': 'The mudpy MUD server engine.',
    'extra_nav_links': collections.OrderedDict((
        ('Browse Source', 'https://mudpy.org/code/mudpy/'),
        ('Bug Reporting', 'https://mudpy.org/bugs/mudpy/'),
        ('Git Clone URL', 'https://mudpy.org/code/mudpy/'),
        ('Release Files', 'https://mudpy.org/dist/mudpy/'),
    )),
    'fixed_sidebar': True,
    'logo_name': True,
    'logo_text_align': 'center',
}
html_title = 'mudpy'
htmlhelp_basename = '%sdoc' % project
latex_documents = [(
    'index', '%s.tex' % project, '%s' % project, 'Jeremy Stanley', 'manual')]
pygments_style = 'sphinx'
source_suffix = '.rst'



"""Data interface functions for the mudpy engine."""

# Copyright (c) 2004-2021 mudpy authors. Permission to use, copy,
# modify, and distribute this software is granted under terms
# provided in the LICENSE file distributed with this software.

import os
import re
import stat

import mudpy
import yaml


class _IBSEmitter(yaml.emitter.Emitter):

    """Override the default YAML Emitter to indent block sequences."""

    def expect_block_sequence(self):
        """Match the expectations of the ``yamllint`` style checker."""

        # TODO(fungi) Get an option for this implemented upstream in
        # the pyyaml library
        self.increase_indent(flow=False, indentless=False)
        self.state = self.expect_first_block_sequence_item


class _IBSDumper(yaml.SafeDumper, _IBSEmitter):

    """Use our _IBSEmitter instead of the default implementation."""

    pass


class Data:

    """A file containing universe elements and their facets."""

    def __init__(self,
                 source,
                 universe,
                 flags=None,
                 relative=None,
                 ):
        self.source = source
        self.universe = universe
        if flags is None:
            self.flags = []
        else:
            self.flags = flags[:]
        self.relative = relative
        self.load()

    def load(self):
        """Read a file, create elements and poplulate facets accordingly."""
        self.modified = False
        self.source = find_file(
                self.source, relative=self.relative, universe=self.universe)
        try:
            with open(self.source) as datafd:
                self.data = yaml.safe_load(datafd)
            log_entry = ("Loaded file %s into memory." % self.source, 5)
        except FileNotFoundError:
            # it's normal if the file is one which doesn't exist yet
            self.data = {}
            log_entry = ("File %s is unavailable." % self.source, 6)
        try:
            mudpy.misc.log(*log_entry)
        except NameError:
            # happens when we're not far enough along in the init process
            self.universe.setup_loglines.append(log_entry)
        if not hasattr(self.universe, "files"):
            self.universe.files = {}
        self.universe.files[self.source] = self
        includes = []
        for node in list(self.data):
            if node == "_load":
                includes += self.data["_load"]
                continue
            if node.startswith("_"):
                continue
            facet_pos = node.rfind(".") + 1
            prefix = node[:facet_pos].strip(".")
            try:
                element = self.universe.contents[prefix]
            except KeyError:
                element = mudpy.misc.Element(prefix, self.universe, self)
            element.set(node[facet_pos:], self.data[node])
            if prefix.startswith("mudpy.movement."):
                self.universe.directions.add(
                    prefix[prefix.rfind(".") + 1:])
        for include_file in includes:
            if not os.path.isabs(include_file):
                include_file = find_file(
                    include_file,
                    relative=self.source,
                    universe=self.universe
                )
            if (include_file not in self.universe.files or not
                    self.universe.files[include_file].is_writeable()):
                Data(include_file, self.universe)

    def save(self):
        """Write the data, if necessary."""
        normal_umask = 0o0022
        private_umask = 0o0077
        private_file_mode = 0o0600

        # when modified, writeable and has content or the file exists
        if self.modified and self.is_writeable() and (
           self.data or os.path.exists(self.source)
           ):

            # make parent directories if necessary
            old_umask = os.umask(normal_umask)
            os.makedirs(os.path.dirname(self.source), exist_ok=True)
            os.umask(old_umask)

            # backup the file
            if "mudpy.limit" in self.universe.contents:
                max_count = self.universe.contents["mudpy.limit"].get(
                    "backups", 0)
            else:
                max_count = 0
            if os.path.exists(self.source) and max_count:
                backups = []
                for candidate in os.listdir(os.path.dirname(self.source)):
                    if re.match(
                       os.path.basename(self.source) +
                       r"""\.\d+$""", candidate
                       ):
                        backups.append(int(candidate.split(".")[-1]))
                backups.sort()
                backups.reverse()
                for old_backup in backups:
                    if old_backup >= max_count - 1:
                        os.remove(self.source + "." + str(old_backup))
                    elif not os.path.exists(
                        self.source + "." + str(old_backup + 1)
                    ):
                        os.rename(
                            self.source + "." + str(old_backup),
                            self.source + "." + str(old_backup + 1)
                        )
                if not os.path.exists(self.source + ".0"):
                    os.rename(self.source, self.source + ".0")

            # our data file
            if "private" in self.flags:
                old_umask = os.umask(private_umask)
                file_descriptor = open(self.source, "w")
                if oct(stat.S_IMODE(os.stat(
                        self.source)[stat.ST_MODE])) != private_file_mode:
                    # if it's marked private, chmod it appropriately
                    os.chmod(self.source, private_file_mode)
            else:
                old_umask = os.umask(normal_umask)
                file_descriptor = open(self.source, "w")
            os.umask(old_umask)

            # write and close the file
            yaml.dump(self.data, Dumper=_IBSDumper, allow_unicode=True,
                      default_flow_style=False, explicit_start=True, indent=4,
                      stream=file_descriptor)
            file_descriptor.close()

            # unset the modified flag
            self.modified = False

    def is_writeable(self):
        """Returns True if the _lock is False."""
        try:
            return not self.data.get("_lock", False)
        except KeyError:
            return True


def find_file(
    file_name=None,
    group=None,
    prefix=None,
    relative=None,
    search=None,
    stash=None,
    universe=None
):
    """Return an absolute file path based on configuration."""

    # this is all unnecessary if it's already absolute
    if file_name and os.path.isabs(file_name):
        return os.path.realpath(file_name)

    # if a universe was provided, try to get some defaults from there
    if universe:

        if hasattr(
                universe, "contents") and "mudpy.filing" in universe.contents:
            filing = universe.contents["mudpy.filing"]
            if not prefix:
                prefix = filing.get("prefix")
            if not search:
                search = filing.get("search")
            if not stash:
                stash = filing.get("stash")

        # if there's only one file loaded, try to work around a chicken<egg
        elif hasattr(universe, "files") and len(
            universe.files
        ) == 1 and not universe.files[
                list(universe.files.keys())[0]].is_writeable():
            data_file = universe.files[list(universe.files.keys())[0]].data

            # try for a fallback default directory
            if not stash:
                stash = data_file.get(".mudpy.filing.stash", "")

            # try for a fallback root path
            if not prefix:
                prefix = data_file.get(".mudpy.filing.prefix", "")

            # try for a fallback search path
            if not search:
                search = data_file.get(".mudpy.filing.search", "")

        # another fallback root path, this time from the universe startdir
        if hasattr(universe, "startdir"):
            if not prefix:
                prefix = universe.startdir
            elif not os.path.isabs(prefix):
                prefix = os.path.join(universe.startdir, prefix)

    # when no root path is specified, assume the current working directory
    if (not prefix or prefix == ".") and hasattr(universe, "startdir"):
        prefix = universe.startdir

    # make sure it's absolute
    prefix = os.path.realpath(prefix)

    # if there's no search path, just use the root path and etc
    if not search:
        search = [prefix, "etc"]

    # work on a copy of the search path, to avoid modifying the caller's
    else:
        search = search[:]

    # if there's no default path, use the last component of the search path
    if not stash:
        stash = search[-1]

    # if an existing file or directory reference was supplied, prepend it
    if relative:
        if os.path.isdir(relative):
            search = [relative] + search
        else:
            search = [os.path.dirname(relative)] + search

    # make the search path entries absolute and throw away any dupes
    clean_search = []
    for each_path in search:
        if not os.path.isabs(each_path):
            each_path = os.path.realpath(os.path.join(prefix, each_path))
        if each_path not in clean_search:
            clean_search.append(each_path)

    # start hunting for the file now
    for each_path in clean_search:

        # construct the candidate path
        candidate = os.path.join(each_path, file_name)

        # if the file exists and is readable, we're done
        if os.path.isfile(candidate):
            file_name = os.path.realpath(candidate)
            break

        # if the path is a directory, look for an __init__ file
        if os.path.isdir(candidate):
            file_name = os.path.realpath(
                    os.path.join(candidate, "__init__.yaml"))
            break

    # it didn't exist after all, so use the default path instead
    if not os.path.isabs(file_name):
        file_name = os.path.join(stash, file_name)
    if not os.path.isabs(file_name):
        file_name = os.path.join(prefix, file_name)

    # and normalize it last thing before returning
    file_name = os.path.realpath(file_name)
    return file_name
